# Use a simple (but explicit) path modification to resolve the package properly
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))
# import sys
# sys.path.append('../')

import pytest
import math
import numpy as np

import autodiffcst.AD as ad
from autodiffcst.AD_vec import *
# from autodiffcst.AD import AD
# from autodiffcst.AD_vec import VAD, jacobian, hessian
import autodiffcst.admath as admath


def test_initialize_base():
    x = VAD(val = [1,2])
    assert np.sum(x.val == np.array([1,2])) == len(x),"Error: initialize x value."
    assert np.sum(x.der == np.array([[1,0],[0,1]])) == 4,"Error: initialize x value."
    assert np.sum(x.der2 == np.array([[[0., 0.],[0., 0.]],[[0., 0.],[0., 0.]]])) == 8,"Error: initialize x value."
    with pytest.raises(TypeError):
        y = VAD([1,2], order=1.2)
    with pytest.raises(ValueError):
        y = VAD([1, 2], order=-1)
    with pytest.raises(Exception):
        y = VAD([1, 2], order=3)



def test_initialize_advanced():
    [x,y] = VAD([1,2])
    assert np.sum(x.val == np.array([1])) == len(x),"Error: initialize x value."
    assert np.sum(x.der == np.array([1., 0.])) == 2,"Error: initialize x value."
    assert np.sum(x.der2 == np.array([[0., 0.],[0., 0.]])) == 4,"Error: initialize x value."

def test_repr():
    x = VAD([1])
    assert repr(x) == "VAD(value: [1], derivatives: [[1.]])", "Error: repr is not working"
    assert x.__repr__() == "VAD(value: [1], derivatives: [[1.]])", "Error: repr is not working"
    assert x.__str__() == "VAD(value: [1], derivatives: [[1.]])", "Error: str is not working"

def test_negative():
    [x,y,z] = VAD([1,2,3])
    x = -x 
    assert np.sum(x.der == np.array([[-1, 0, 0]])) == 3, "Error: AD first derivative of negation is wrong."
    assert np.sum(x.der2 == np.array([[0., 0., 0.], [0., 0., 0.],[0., 0., 0.]])) == 9, "Error: AD second derivative of negation is wrong"

    f = VAD([3,4,5])
    f = -f
    assert np.sum(f.der == np.array([[-1, 0, 0], [0, -1, 0], [0, 0, -1]])) == 9, "Error: VAD first derivative of negation is wrong."
    assert np.sum(f.der2 == np.array([[[0., 0., 0.], [0., 0., 0.],[0., 0., 0.]]])) == 27, "Error: VAD second derivative of negation is wrong"

def test_eq_VAD():
    A = VAD([1,2,3])
    B = VAD([1,2,1])
    eq1 = A == B
    eq2 = A == A
    assert eq1 == False, "Error: dunder equal for VAD is wrong."
    assert eq2 == True, "Error: dunder equal for VAD is wrong."

def test_eq_AD():
    A = VAD([1,2,3])
    B = VAD([1,2,1])
    eq1 = A[0] == B[2]
    eq2 = A[0] == A[2]
    assert eq1 == True, "Error: dunder equal for VAD is wrong."
    assert eq2 == False, "Error: dunder equal for VAD is wrong."
    with pytest.raises(TypeError):
        A == 1
        A > 1
        A < 1
        A >= 1
        A <= 1
    
def test_len_VAD():
    A = VAD([1,2,3])
    l = len(A)
    assert l == 3, "Error: dunder length for VAD is wrong."
    
def test_len_AD():
    [x, y] = VAD([1,2])
    lx = len(x)
    assert lx == 1, "Error: dunder length for AD is wrong."
    

def test_isequal_VAD():
    A = VAD([1,2,3])
    B = VAD([1,2,1])
    eq1 = A.isequal(B)
    eq2 = A.isequal(A)
    assert np.sum(eq1 == np.array([True, True, False])) == 3, "Error: isequal for VAD is wrong."
    assert np.sum(eq2 == np.array([True, True, True])) == 3, "Error: isequal for VAD is wrong."

def test_add_AD():
    # x = ad.AD(3)
    # y = ad.AD(1)
    [x, y] = VAD([3,1])
    f = x + y
    g = x + 1
    f += y
    
    assert f.val == 5, "Error: add value for AD is wrong."
    assert np.sum(f.der == np.array([1, 2])) == 2, "Error: add der for AD is wrong."
    
    assert np.sum(f.der2 == np.array([[0, 0],[0, 0]])) == 4, "Error: add der2 for AD is wrong."
    assert f-1 == g, "Error: add comparison for AD is wrong."
    
    # assert f.val == 4, "Error: add value for AD is wrong."
    # assert np.sum(f.der == np.array([1, 1])) == 2, "Error: add der for AD is wrong."
    

def test_add_VAD():
    
    f = VAD([2,1])
    m = f + 1
    g = 1 + m
    g += f
    
    h = VAD([0,0]) + VAD([6,4])
    assert np.sum(g.val == np.array([6, 4])) == 2, "Error: add value for VAD is wrong."
    assert np.sum(g.der == np.array([[2, 0],[0,2]])) == 4, "Error: add der for VAD is wrong."
    
    assert np.sum(g.der2 == np.array([[[0, 0],[0, 0]],[[0, 0],[0, 0]]])) == 8, "Error: add der2 for VAD is wrong."
    
    assert g == h, "Error: add value for VAD is wrong."

def test_sub_AD():
    # x = ad.AD(3)
    # y = ad.AD(1)
    [x, y] = VAD([3,1])
    f = 0 - (x - y)
    assert f.val == -2, "Error: truediv value for AD is wrong."
    assert np.sum(f.der == np.array([-1, 1])) == 2, "Error: truediv der for AD is wrong."
    assert np.sum(f.der2 == np.array([[0, 0],[0, 0]])) == 4, "Error: truediv der for AD is wrong."


def test_sub_VAD():
    f = VAD([3, 1])
    g = f - 4
    h = VAD([1, 1]) - VAD([2, 4])
    assert np.sum(g.val == np.array([-1, -3])) == 2, "Error: sub value for VAD is wrong."
    assert np.sum(g.der == np.array([[1, 0], [0, 1]])) == 4, "Error: sub der for VAD is wrong."

    assert np.sum(g.der2 == np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])) == 8, "Error: sub der2 for VAD is wrong."

    assert g == h, "Error: sub value for VAD is wrong."

def test_mul_AD():
    [x, y] = VAD([3,1])
    f = x * y
    g = 4 * x
    g *= y
    h = g*y

    assert f.val == 3, "Error: mul value for AD is wrong."
    assert np.sum(f.der == np.array([1, 3])) == 2, "Error: mul der for AD is wrong."
    assert np.sum(f.der2 == np.array([[0, 1],[1, 0]])) == 4, "Error: mul der for AD is wrong."
    
    assert g.val == 12, "Error: mul value for AD is wrong."
    assert np.sum(g.der == np.array([4, 12])) == 2, "Error: mul der for AD is wrong."
    assert np.sum(g.der2 == np.array([[0, 4], [4, 0]])) == 4, "Error: mul der for AD is wrong."
    
    assert h.val == 12, "Error: mul value for AD is wrong."
    assert np.sum(h.der == np.array([4, 24])) == 2, "Error: mul der for AD is wrong."
    assert np.sum(h.der2 == np.array([[0, 8], [8, 24]])) == 4, "Error: mul der for AD is wrong."

def test_mul_VAD():
    f = VAD([3,1])
    g = 4*f
    g *= 2
    assert np.sum(g.val == np.array([24, 8])) == 2, "Error: mul value for VAD is wrong."
    assert np.sum(g.der == np.array([[8, 0], [0, 8]])) == 4, "Error: mul der for VAD is wrong."
    assert np.sum(g.der2 == np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])) == 8, "Error: mul der2 for VAD is wrong."

def test_div_AD():
    [x, y] = VAD([3,1])
    f = x / y
    g = 9 / x
    g /= 1
    assert f.val == 3, "Error: div value for AD is wrong."
    assert np.sum(f.der == np.array([1, -3])) == 2, "Error: div der for AD is wrong."
    assert np.sum(f.der2 == np.array([[0, -1],[-1, 6]])) == 4, "Error: div der for AD is wrong."

    assert g.val == 3, "Error: div value for AD is wrong."
    assert np.sum(g.der == np.array([-1, 0])) == 2, "Error: div der for AD is wrong."
    assert np.sum(g.der2 == np.array([[18/27, 0], [0, 0]])) == 4, "Error: div der for AD is wrong."

def test_div_VAD():
    f = VAD([3,1])
    h = f / 3
    g = 1 / f
    g /= 1
    assert np.sum(h.val == np.array([1, 1/3])) == 2, "Error: div value for AD is wrong."
    assert np.sum(h.der == np.array([[1/3, 0], [0, 1/3]])) == 4, "Error: div der for AD is wrong."
    assert np.sum(h.der2 == np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])) == 8, "Error: div der for AD is wrong."
    assert np.sum(g.val == np.array([1/3, 1])) == 2, "Error: div value for AD is wrong."
    assert np.sum(g.der == np.array([[-1/9, 0], [0, -1]])) == 4, "Error: div der for AD is wrong."
    assert np.sum(g.der2 == np.array([[[2/27, 0], [0, 0]], [[0, 0], [0, 2]]])) == 8, "Error: div der for AD is wrong."

def test_pow_AD():
    [x, y] = VAD([3,1])
    f = x ** 2
    assert f.val == 9, "Error: pow value for AD is wrong."
    assert np.sum(f.der == np.array([6, 0])) == 2, "Error: pow der for AD is wrong."
    assert np.sum(f.der2 == np.array([[2, 0],[0, 0]])) == 4, "Error: pow der for AD is wrong."
    g = 2 ** x
    # precision is different from np
    assert np.allclose(g.val, np.array([8.0])), "Error: pow value for AD is wrong."
    assert np.allclose(g.der,np.array([8*np.log(2), 0.0])), "Error: pow der for AD is wrong."
    assert np.allclose(g.der2, np.array([[8*np.log(2)*np.log(2), 0.0], [0., 0.]])), "Error: pow der for AD is wrong."

def test_set_VAD():
    x = ad.AD(1, tag=0, size=2)
    y = ad.AD(2, tag=1, size=2)
    ADs = np.array([x, y])
    assert set_VAD(ADs) == VAD([1,2]), "Error: set_VAD is wrong."

def test_decorator():
    func = my_decorator(admath.exp)
    y = VAD([1, 2, 3])
    assert func(y) == exp(y)

def test_jacobian():
    x = VAD([3, 1])
    f = 2 * x
    g = 2 * x[0]
    assert np.allclose(jacobian(f),np.array([[2., 0.],[0.,2.]]))
    assert np.allclose(jacobian(g),np.array([2., 0.]))

def test_hessian():
    x = VAD([3, 1])
    g = 2 * x[0]
    assert np.allclose(hessian(g), np.array([[0., 0.], [0., 0.]]))


def test_pow_m():
    x = VAD([1, 2])
    g = pow(x, 2)
    assert np.allclose(g.val, np.array([1., 4.]))
    assert np.allclose(g.der, np.array([[2., 0.],[0., 4.]]))

def test_diff():
    x = VAD([3, 1])
    f = 2 * x
    g = x[1] * x[0]
    assert np.allclose(f.diff(1, 1), np.array([0.,2.]))
    assert np.allclose(f.diff([1, 0], 2), np.array([0.,0.]))
    assert g.diff(0, 1) == 1.0
    assert g.diff(1, 1) == 3.0
    assert g.diff([0, 0], 2) ==0.0
    with pytest.raises(Exception):
        g.diff(0, 2)


def test_fullequal():
    a = VAD([1, 2, 3])
    b = VAD([1, 2, 3]) * 2
    assert np.sum(a.fullequal(b)) ==0
    assert a[0].fullequal(a[0])
    with pytest.raises(TypeError):
        a.fullequal(1)

def test_gt():
    a = VAD([1, 2, 3])
    b = VAD([2, 2, 3])
    assert not a > b
    assert b[0] > a[0]
    with pytest.raises(TypeError):
        assert a > 1

def test_ge():
    a = VAD([1, 2, 3])
    b = VAD([2, 2, 3])
    assert not a >= b
    assert b[0] >= a[0]
    with pytest.raises(TypeError):
        assert a>= 1


def test_isgreater():
    a = VAD([1, 3, 3])
    b = VAD([2, 2, 3])
    assert np.sum(a.isgreater(b) == np.array([False, True, False])) == 3
    with pytest.raises(TypeError):
        a.isgreater(2)

def test_le():
    a = VAD([1, 1, 1])
    b = VAD([2, 2, 3])
    assert a <= b
    assert not b[0] <= a[0]
    with pytest.raises(TypeError):
        assert a <= 1

def test_lt():
    a = VAD([1, 2, 3])
    b = VAD([2, 2, 3])
    assert not a < b
    assert not b[0] <= a[0]
    with pytest.raises(TypeError):
        assert a < 1

def test_isless():
    a = VAD([1, 2, 3])
    b = VAD([2, 2, 3])
    assert np.sum(a.isless(b) == np.array([True, False, False])) ==3
    with pytest.raises(TypeError):
        a.isless(3)

def test_sub_m():
    a = VAD([1])
    b = 1 - a
    b -= 1
    assert b.val == -1.0
    assert b.der == -1.0


def test_mod():
    a = VAD([10])
    assert a%3 == 1
    with pytest.raises(TypeError):
        assert a%[1,2] ==1

def test_AD_init_hw():
    with pytest.raises(TypeError):
        x = ad.AD(-1,tag=0,size=1.5)
        x = ad.AD(-1,tag=0,order=1.5)
        x = ad.AD(-1,tag=0,order='error')
        x = VAD([-1],order='error')
    with pytest.raises(ValueError):
        x = ad.AD(-1,order=0)
    with pytest.raises(Exception):
        x = VAD([1,2,3],order=5)
        x = ad.AD([1,2,3],order=5)

def test_AD_repr_hw():
    x = ad.AD([1],tag=0,der=[1],order=5)
    assert x.__repr__() == 'AD(value: [[1]], derivatives: [1.])', "Error: repr is not working"
    assert x.__str__() == 'AD(value: [[1]], derivatives: [1.])', "Error: str is not working"

def test_AD_eq_hw():
    x = ad.AD([1],tag=0,der=[1],order=5)
    y = VAD([1])
    with pytest.raises(TypeError):
        x == y
        x.fullequal(y)
        x < y
        x > y
        x <= y
        x >= y
        y = VAD([1])
    
    y = 0
    with pytest.raises(TypeError):
        x == y
        x.fullequal(y)
        x < y
        x > y
        x <= y
        x >= y

def test_AD_ne_hw():
    x = ad.AD([1],tag=0,der=[1],order=5)
    assert not x != x, "Error: ne not working for AD"

def test_AD_ueq_hw():
    x = ad.AD([1],tag=0,der=[1],order=5)
    y = 2*x
    assert x != y, "Error: ne not working for AD"
    assert x < y, "Error: ne not working for AD"
    assert y > x, "Error: ne not working for AD"
    assert x <= y, "Error: ne not working for AD"
    assert y >= x, "Error: ne not working for AD"

def test_AD_add_hw():
    x = ad.AD([1],tag=0,der=[1],order=5)
    y = ad.AD([1],tag=0,der=[1],order=7)
    f = x + x
    with pytest.raises(Exception):
        f = x + y

def test_higher_hw():
    x = ad.AD(val = 1, order = 10, size = 1, tag = 0)
    f = x**5
    assert f.higherdiff(1) == 5,"higherdiff not working"
    with pytest.raises(TypeError):
        f.higherdiff('error')
    with pytest.raises(ValueError):
        f.higherdiff(0)
        f.higherdiff(15)
    with pytest.raises(Exception):
        x = ad.AD(val = 1, size = 1, tag = 0)
        f = x**5
        f.higherdiff(10)

def test_AD_rpow_hw():
    x = ad.AD(val = 1, order = 10, size = 1, tag = 0)
    f = 5**x
    assert np.abs(f.val - 5)<1e-8, "rpow not working for AD"

def test_AD_pow_hw():
    x = ad.AD(val = 1, order = 10, size = 1, tag = 0)
    f = x**5

    assert np.abs(f.val - 1)<1e-8, "rpow not working for AD"
    x = ad.AD(val = 0, order = 10, size = 1, tag = 0)
    
    with pytest.raises(ValueError):
        f = x ** 5
        f = x**2
    with pytest.raises(TypeError):
        f = x**'a'
    
def test_VAD_op_hw():
    x = VAD(val = [1,2])
    y = VAD(val = [1,2])
    z = VAD(val = [1,2])
    x -= 1
    assert np.allclose(x.val,np.array([0,1])), 'Error: wrong isub VAD'
    y %= 2
    assert np.allclose(y,np.array([1,0])), 'Error: wrong imod VAD'
    z *= 2
    assert np.allclose(z.val,np.array([2,4])), 'Error: wrong imul VAD'
    x = VAD(val = [2,2])
    x /= 2
    assert np.allclose(x.val,np.array([1,1])), 'Error: wrong itruediv VAD'
    x = VAD(val = [2,2])
    f = x**2
    assert np.allclose(f.val,np.array([4,4])), 'Error: wrong pow VAD'


# -*- coding: utf-8 -*-

"""
Integration part for hooking into distutils/setuptools

Rationale:
The ``use_pyscaffold`` keyword is unknown to setuptools' setup(...) command,
therefore the ``entry_points`` are checked for a function to handle this
keyword which is ``pyscaffold_keyword`` below. This is where we hook into
setuptools and apply the magic of setuptools_scm as well as other commands.
"""
from distutils.cmd import Command

from .contrib import ptr
from .contrib.setuptools_scm.integration import version_keyword
from .repo import get_git_root
from .utils import check_setuptools_version


def version2str(version):
    """Creates a PEP440 version string

    Args:
        version (:obj:`setuptools_scm.version.ScmVersion`): version object

    Returns:
        str: version string
    """
    if version.exact or not version.distance > 0:
        return version.format_with('{tag}')
    else:
        distance = version.distance
        version = str(version.tag)
        if '.dev' in version:
            version, tail = version.rsplit('.dev', 1)
            assert tail == '0', 'own dev numbers are unsupported'
        return '{}.post0.dev{}'.format(version, distance)


def local_version2str(version):
    """Create the local part of a PEP440 version string

    Args:
        version (:obj:`setuptools_scm.version.ScmVersion`): version object

    Returns:
        str: local version
    """
    if version.exact:
        return ''
    else:
        if version.dirty:
            return version.format_with('+{node}.dirty')
        else:
            return version.format_with('+{node}')


def setuptools_scm_config(value):
    """Generate the configuration for setuptools_scm

    Args:
        value: value from entry_point

    Returns:
        dict: dictionary of options
    """
    value = value if isinstance(value, dict) else dict()
    value.setdefault('root', get_git_root(default='.'))
    value.setdefault('version_scheme', version2str)
    value.setdefault('local_scheme', local_version2str)
    return value


def build_cmd_docs():
    """Return Sphinx's BuildDoc if available otherwise a dummy command

    Returns:
        :obj:`~distutils.cmd.Command`: command object
    """
    try:
        from sphinx.setup_command import BuildDoc
    except ImportError:
        class NoSphinx(Command):
            user_options = []

            def initialize_options(self):
                raise RuntimeError("Sphinx documentation is not installed, "
                                   "run: pip install sphinx")

        return NoSphinx
    else:
        return BuildDoc


def pyscaffold_keyword(dist, keyword, value):
    """Handles the `use_pyscaffold` keyword of the setup(...) command

    Args:
        dist (:obj:`setuptools.dist`): distribution object as
        keyword (str): keyword argument = 'use_pyscaffold'
        value: value of the keyword argument
    """
    check_setuptools_version()
    if value:
        version_keyword(dist, keyword, setuptools_scm_config(value))
        dist.cmdclass['docs'] = build_cmd_docs()
        dist.cmdclass['doctest'] = build_cmd_docs()
        dist.cmdclass['build_sphinx'] = build_cmd_docs()
        dist.command_options['doctest'] = {'builder': ('setup.py', 'doctest')}
        dist.cmdclass['test'] = ptr.PyTest



# -*- coding: utf-8 -*-
"""
Extension that generates configuration files for the Tox test automation tool.
"""

from ..api import Extension, helpers
from ..templates import tox as tox_ini


class Tox(Extension):
    """Generate Tox configuration file"""
    def activate(self, actions):
        """Activate extension

        Args:
            actions (list): list of actions to perform

        Returns:
            list: updated list of actions
        """
        return self.register(
            actions,
            self.add_files,
            after='define_structure')

    def add_files(self, struct, opts):
        """Add .tox.ini file to structure

        Args:
            struct (dict): project representation as (possibly) nested
                :obj:`dict`.
            opts (dict): given options, see :obj:`create_project` for
                an extensive list.

        Returns:
            struct, opts: updated project representation and options
        """
        files = {
            'tox.ini': (tox_ini(opts), helpers.NO_OVERWRITE)
        }

        return helpers.merge(struct, {opts['project']: files}), opts



# -*- coding: utf-8 -*-
"""
Warnings used by PyScaffold to identify issues that can be safely ignored
but that should be displayed to the user.
"""


class UpdateNotSupported(RuntimeWarning):
    """Extensions that make use of external generators are not able to do
    updates by default.
    """

    DEFAULT_MESSAGE = ('Updating code generated using external tools is not '
                       'supported. The extension `{}` will be ignored, only '
                       'changes in PyScaffold core features will take place.')

    def __init__(self, *args, extension=None, **kwargs):
        if not args:
            args = [self.DEFAULT_MESSAGE.format(extension)]
        super(UpdateNotSupported, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
Extension that creates a base structure for the project using django-admin.py.

Warning:
    *Deprecation Notice* - In the next major release the Django extension
    will be extracted into an independent package.
    After PyScaffold v4.0, you will need to explicitly install
    ``pyscaffoldext-django`` in your system/virtualenv in order to be
    able to use it.
"""
import os
import shutil
from os.path import join as join_path

from .. import shell
from ..api import Extension, helpers
from ..warnings import UpdateNotSupported


class Django(Extension):
    """Generate Django project files"""
    mutually_exclusive = True

    def activate(self, actions):
        """Register hooks to generate project using django-admin.

        Args:
            actions (list): list of actions to perform

        Returns:
            list: updated list of actions
        """

        # `get_default_options` uses passed options to compute derived ones,
        # so it is better to prepend actions that modify options.
        actions = helpers.register(actions, enforce_django_options,
                                   before='get_default_options')
        # `apply_update_rules` uses CWD information,
        # so it is better to prepend actions that modify it.
        actions = helpers.register(actions, create_django_proj,
                                   before='apply_update_rules')

        return actions


def enforce_django_options(struct, opts):
    """Make sure options reflect the Django usage.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`.
        opts (dict): given options, see :obj:`create_project` for
            an extensive list.

    Returns:
        struct, opts: updated project representation and options
    """
    opts['package'] = opts['project']  # required by Django
    opts['force'] = True
    opts.setdefault('requirements', []).append('django')

    return struct, opts


def create_django_proj(struct, opts):
    """Creates a standard Django project with django-admin.py

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`.
        opts (dict): given options, see :obj:`create_project` for
            an extensive list.

    Returns:
        struct, opts: updated project representation and options

    Raises:
        :obj:`RuntimeError`: raised if django-admin.py is not installed
    """
    if opts.get('update'):
        helpers.logger.warning(UpdateNotSupported(extension='django'))
        return struct, opts

    try:
        shell.django_admin('--version')
    except Exception as e:
        raise DjangoAdminNotInstalled from e

    pretend = opts.get('pretend')
    shell.django_admin('startproject', opts['project'],
                       log=True, pretend=pretend)
    if not pretend:
        src_dir = join_path(opts['project'], 'src')
        os.mkdir(src_dir)
        shutil.move(join_path(opts['project'], opts['project']),
                    join_path(src_dir, opts['package']))

    return struct, opts


class DjangoAdminNotInstalled(RuntimeError):
    """This extension depends on the ``django-admin.py`` cli script."""

    DEFAULT_MESSAGE = ("django-admin.py is not installed, "
                       "run: pip install django")

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super(DjangoAdminNotInstalled, self).__init__(message, *args, **kwargs)



from __future__ import print_function
import sys
from setuptools_scm import get_version
from setuptools_scm.integration import find_files
from setuptools_scm.version import _warn_if_setuptools_outdated


def main():
    _warn_if_setuptools_outdated()
    print("Guessed Version", get_version())
    if "ls" in sys.argv:
        for fname in find_files("."):
            print(fname)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Functionality to update one PyScaffold version to another
"""
import os
from functools import reduce
from os.path import exists as path_exists
from os.path import join as join_path

from pkg_resources import parse_version

from . import __version__ as pyscaffold_version
from .contrib.configupdater import ConfigUpdater
from .log import logger
from .structure import FileOp
from .utils import get_id, get_setup_requires_version


def apply_update_rules(struct, opts, prefix=None):
    """Apply update rules using :obj:`~.FileOp` to a directory structure.

    As a result the filtered structure keeps only the files that actually will
    be written.

    Args:
        opts (dict): options of the project, containing the following flags:

            - **update**: when the project already exists and should be updated
            - **force**: overwrite all the files that already exist

        struct (dict): directory structure as dictionary of dictionaries
            (in this tree representation, each leaf can be just a
            string or a tuple also containing an update rule)
        prefix (str): prefix path for the structure

    Returns:
        tuple(dict, dict):
            directory structure with keys removed according to the rules
            (in this tree representation, all the leaves are strings) and input
            options
    """
    if prefix is None:
        prefix = os.getcwd()

    filtered = {}

    for k, v in struct.items():
        if isinstance(v, dict):
            v, _ = apply_update_rules(v, opts, join_path(prefix, k))
        else:
            path = join_path(prefix, k)
            v = apply_update_rule_to_file(path, v, opts)

        if v is not None:
            filtered[k] = v

    return filtered, opts


def apply_update_rule_to_file(path, value, opts):
    """Applies the update rule to a given file path

    Args:
        path (str): file path
        value (tuple or str): content (and update rule)
        opts (dict): options of the project, containing the following flags:

            - **update**: if the project already exists and should be updated
            - **force**: overwrite all the files that already exist

    Returns:
        content of the file if it should be generated or None otherwise.
    """
    if isinstance(value, (tuple, list)):
        content, rule = value
    else:
        content, rule = value, None

    update = opts.get('update')
    force = opts.get('force')

    skip = update and not force and (
            rule == FileOp.NO_CREATE or
            path_exists(path) and rule == FileOp.NO_OVERWRITE)

    if skip:
        logger.report('skip', path)
        return None

    return content


def read_setupcfg(project_path):
    """Reads-in setup.cfg for updating

    Args:
        project_path (str): path to project

    Returns:

    """
    path = join_path(project_path, 'setup.cfg')
    updater = ConfigUpdater()
    updater.read(path, encoding='utf-8')
    return updater


def invoke_action(action, struct, opts):
    """Invoke action with proper logging.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`.
        opts (dict): given options, see :obj:`create_project` for
            an extensive list.

    Returns:
        tuple(dict, dict): updated project representation and options
    """
    logger.report('invoke', get_id(action))
    with logger.indent():
        struct, opts = action(struct, opts)

    return struct, opts


def get_curr_version(project_path):
    """Retrieves the PyScaffold version that put up the scaffold

    Args:
        project_path: path to project

    Returns:
        Version: version specifier
    """
    setupcfg = read_setupcfg(project_path).to_dict()
    return parse_version(setupcfg['pyscaffold']['version'])


def version_migration(struct, opts):
    """Migrations from one version to another

    Args:
        struct (dict): previous directory structure (ignored)
        opts (dict): options of the project

    Returns:
        tuple(dict, dict):
            structure as dictionary of dictionaries and input options
    """
    update = opts.get('update')

    if not update:
        return struct, opts

    curr_version = get_curr_version(opts['project'])

    # specify how to migrate from one version to another as ordered list
    migration_plans = [
        (parse_version('3.1'), [add_entrypoints,
                                add_setup_requires])
    ]
    for plan_version, plan_actions in migration_plans:
        if curr_version < plan_version:
            struct, opts = reduce(lambda acc, f: invoke_action(f, *acc),
                                  plan_actions, (struct, opts))

    # note the updating version in setup.cfg for future use
    update_pyscaffold_version(opts['project'], opts['pretend'])
    # replace the old version with the updated one
    opts['version'] = pyscaffold_version
    return struct, opts


def add_entrypoints(struct, opts):
    """Add [options.entry_points] to setup.cfg

    Args:
        struct (dict): previous directory structure (ignored)
        opts (dict): options of the project

    Returns:
        tuple(dict, dict):
            structure as dictionary of dictionaries and input options
    """
    setupcfg = read_setupcfg(opts['project'])
    section_str = """[options.entry_points]
# Add here console scripts like:
# console_scripts =
#     script_name = ${package}.module:function
# For example:
# console_scripts =
#     fibonacci = ${package}.skeleton:run
# And any other entry points, for example:
# pyscaffold.cli =
#     awesome = pyscaffoldext.awesome.extension:AwesomeExtension
    """
    new_section_name = 'options.entry_points'
    if new_section_name in setupcfg:
        return struct, opts

    new_section = ConfigUpdater()
    new_section.read_string(section_str)
    new_section = new_section[new_section_name]

    add_after_sect = 'options.extras_require'
    if add_after_sect not in setupcfg:
        # user removed it for some reason, default to metadata
        add_after_sect = 'metadata'

    setupcfg[add_after_sect].add_after.section(new_section).space()
    if not opts['pretend']:
        setupcfg.update_file()
    return struct, opts


def add_setup_requires(struct, opts):
    """Add `setup_requires` in setup.cfg

    Args:
        struct (dict): previous directory structure (ignored)
        opts (dict): options of the project

    Returns:
        tuple(dict, dict):
            structure as dictionary of dictionaries and input options
    """
    setupcfg = read_setupcfg(opts['project'])
    comment = ("# DON'T CHANGE THE FOLLOWING LINE! "
               "IT WILL BE UPDATED BY PYSCAFFOLD!")
    options = setupcfg['options']
    if 'setup_requires' in options:
        return struct, opts

    version_str = get_setup_requires_version()
    (options['package_dir'].add_after
                           .comment(comment)
                           .option('setup_requires', version_str))
    if not opts['pretend']:
        setupcfg.update_file()
    return struct, opts


def update_pyscaffold_version(project_path, pretend):
    """Update `setup_requires` in setup.cfg

    Args:
        project_path (str): path to project
        pretend (bool): only pretend to do something
    """
    setupcfg = read_setupcfg(project_path)
    setupcfg['options']['setup_requires'] = get_setup_requires_version()
    setupcfg['pyscaffold']['version'] = pyscaffold_version
    if not pretend:
        setupcfg.update_file()



import math
import numbers
import numpy as np
import warnings
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from autodiffcst.admath import chain_rule,fact_ad,choose, log, exp
# from admath import chain_rule,fact_ad,choose

class AD():

    def __init__(self, val, order=2, size = None, tag=None, der=None, der2=None,  higher=None): 
        """
        Overwrites the __init__ dunder method to create a new AD object with initial value and derivatives.
    
                Parameters:
                        val (int or float): the initial value of the new AD object.
                        order (int): the highest order of derivatives the user wants to evaluate
                        size (int): the size of dimension that the AD object resides
                        tag (list of int or np.ndarray of int): the tag, or direction that the AD object resides in its dimension
                        der (list of float or np.ndarray of float): first order derivative of the new AD object, 
                                                                    contained in a list or or np.ndarray           
                        der2 (list of float or np.ndarray of float): second order derivative of the new AD object, 
                                                                    contained in a list or or np.ndarray
                        higher (list of float or np.ndarray of float): higher order derivatives of the new AD object, 
                                                                    contained in a list or or np.ndarray

                Returns:
                        None, but initializes an AD object when called

                Example:
                >>> AD(val=1, order=2, size=1, tag=1) 
                AD(value: [1], derivatives: [1.])

                >>> AD(1, 2, 1, 1) 
                AD(value: [1], derivatives: [1.])

                Note: Initializing val, order, size and tag is a must to use AD directly
        """
        self.val = val if isinstance(val, np.ndarray) else np.array([val])
        if der is None:
            if size is None:
                self.size = 1
                warnings.warn("Size is not specified. The default size is set to 1.", Warning)
            elif isinstance(size, numbers.Integral):
                self.size = size 
            else:
                raise TypeError("Invalid size type. The size of AD can only be integers.")
        else:
            self.size = len(der)


        self.tag = tag if isinstance(tag, np.ndarray) else np.array([tag])
        
        if isinstance(der, np.ndarray):
            self.der = der
        else:
            self.der = np.zeros(self.size)
            self.der[tag] = 1
        
        if isinstance(der2, np.ndarray):
            self.der2 = der2
        else:
            self.der2 = np.zeros((self.size,self.size))
        
        self.order = order
        # self.higher= higher
        if isinstance(order, numbers.Integral):
            if len(self.val) == 1 and order > 2:
                if higher is None:
                    self.higher = np.array([0.0] * order)
                    self.higher[0] = self.der
                    self.higher[1] = self.der2
                else:
                    #print("reach here")
                    self.higher = higher
            elif order > 2:
                raise Exception("Cannot handle higher order derivatives for multiple variables")    
            elif order > 0:
                # must always have higher attribute
                self.higher = None
            else:
                raise ValueError("Order of derivative must be at least 1.")
                
        else:
            raise TypeError("Invalid input for order of derivative.")


    def __repr__(self):
        """
        Overwrites the __repr__ dunder method to nicely print an AD object.
    
                Parameters:
                        self (AD): the AD object that __repr__ is called upon.
    
                Returns:
                        A string containing the current value and derivatives of the AD object.

                Example:
                >>> repr(AD(1, 2, 1, 0))
                "AD(value: [1], derivatives: [1.])"
        """
        return "AD(value: {0}, derivatives: {1})".format(self.val, self.der)

    def __str__(self):
        """
        Overwrites the __str__ dunder method to nicely turn an AD object into a string.
    
                Parameters:
                        self (AD): the AD object that __str__ is called upon.
    
                Returns:
                        A string containing the current value and derivatives of the AD object.

                Example:
                >>> str(AD(1, 2, 1, 0))
                "AD(value: [1], derivatives: [1.])"
        """        
        return "AD(value: {0}, derivatives: {1})".format(self.val,self.der)

    def __eq__(self, other):
        """
        Overwrites the __eq__ dunder method to check if two AD objects are equal in value.
    
                Parameters:
                        self (AD): the AD object that __eq__ is called upon.
                        other (AD): the AD object to be compared with.
    
                Returns:
                        True if the two AD objects have equal values.
                        False if the two AD objects do not have equal values.

                Example:
                >>> AD(1, 2, 1, 0) == AD(2, 2 ,1, 0)
                False
        """  
        if isinstance(other, AD):
            if self.val == other.val: 
                return True
            else:
                return False
        else:
            raise TypeError("Invalid Comparison. AD object can only be compared with AD.")
    
    def __ne__(self, other):
        """
        Overwrites the __ne__ dunder method to check if two AD objects are not equal in value.
    
                Parameters:
                        self (AD): the AD object that __ne__ is called upon.
                        other (AD): the AD object to be compared with.
    
                Returns:
                        False if the two AD objects have equal values.
                        True if the two AD objects do not have equal values.

                Example:
                >>> AD(1, 2, 1, 0) != AD(2, 2 , 1, 0)
                True
        """ 
        return not self == other
    
    def fullequal(self, other):
        """
        Check if two AD objects are equal in value, first and second derivatives.
    
                Parameters:
                        self (AD): the AD object that fullequal is called upon.
                        other (AD): the AD object to be compared with.
    
                Returns:
                        True if the two AD objects have equal values, first and second derivatives.
                        False if the two AD objects differ in any of their equal values, first and second derivatives.

                Example:
                >>> a = AD(1, 2, 1, 0)
                >>> a.fullequal(AD(2, 2 ,1, 0))
                False
        """ 
        if isinstance(other, AD):
            return np.allclose(self.val, other.val) and np.allclose(self.der,other.der) and np.allclose(self.der2, other.der2)
        else:
            raise TypeError("Invalid Comparison. AD object can only be compared with AD.")


    def __lt__(self, other):
        """
        Overwrites the __lt__ dunder method to check if the current AD objects is less than another AD in value.
    
                Parameters:
                        self (AD): the AD object that __lt__ is called upon.
                        other (AD): the AD object to be compared with.
    
                Returns:
                        True if the current AD objects is less than the other AD in value.
                        False if the current AD objects is not less than the other AD in value.

                Example:
                >>> AD(1, 2, 1, 0) < AD(2, 2 ,1, 0)
                array([ True])
        """  
        if isinstance(other, AD):
            return self.val < other.val
        else:
            raise TypeError("Invalid Comparison. AD object can only be compared with AD.")

    def __gt__(self, other):
        """
        Overwrites the __gt__ dunder method to check if the current AD objects is greater than another AD in value.
    
                Parameters:
                        self (AD): the AD object that __gt__ is called upon.
                        other (AD): the AD object to be compared with.
    
                Returns:
                        True if the current AD objects is greater than the other AD in value.
                        False if the current AD objects is not greater than the other AD in value.

                Example:
                >>> AD(1, 2, 1, 0) > AD(2, 2 ,1, 0)
                array([False])
        """ 
        if isinstance(other, AD):
            return self.val > other.val
        else:
            raise TypeError("Invalid Comparison. AD object can only be compared with AD.")

    def __le__(self, other):
        """
        Overwrites the __le__ dunder method to check if the current AD objects is no larger than another AD in value.
    
                Parameters:
                        self (AD): the AD object that __le__ is called upon.
                        other (AD): the AD object to be compared with.
    
                Returns:
                        True if the current AD objects is no larger than the other AD in value.
                        False if the current AD objects is larger than the other AD in value.

                Example:
                >>> AD(1, 2, 1, 0) <= AD(2, 2 ,1, 0)
                array([ True])
        """ 
        if isinstance(other, AD):
            return self.val <= other.val
        else:
            raise TypeError("Invalid Comparison. AD object can only be compared with AD.")


    def __ge__(self, other):
        """
        Overwrites the __ge__ dunder method to check if the current AD objects is no less than another AD in value.
    
                Parameters:
                        self (AD): the AD object that __ge__ is called upon.
                        other (AD): the AD object to be compared with.
    
                Returns:
                        True if the current AD objects is no less than the other AD in value.
                        False if the current AD objects is less than the other AD in value.

                Example:
                >>> AD(1, 2, 1, 0) >= AD(2, 2 ,1, 0)
                array([False])
        """  
        if isinstance(other, AD):
            return self.val >= other.val
        else:
            raise TypeError("Invalid Comparison. AD object can only be compared with AD.")


    def __len__(self):
        """
        Overwrites the __len__ dunder method to get the length of the vector dimension that the AD object resides in.
    
                Parameters:
                        self (AD): the AD object that __len__ is called upon.
    
                Returns:
                        An integer representing the length of the vector dimension that the AD object resides in.

                Example:
                >>> len(AD(1, 2, 1, 0))
                1
        """ 
        return len(self.tag)
    ## Unary 
    def __neg__(self):
        """
        Overwrites the __neg__ dunder method to get the negation of the AD object.
    
                Parameters:
                        self (AD): the AD object that __neg__ is called upon.
    
                Returns:
                        A new AD object which has the negated value and derivative of the current AD.

                Example:
                >>> -AD(1, 2, 1, 0)
                AD(value: [-1], derivatives: [-1.])
        """
        return self*(-1)
    ## Addition
    def __add__(self, other):
        """
        Overwrites the __add__ dunder method to apply addition to an AD object.
    
                Parameters:
                        self (AD): An AD object to be applied addition to
                        other (AD or int or float): the object to be added to self
    
                Returns:
                        new_self (AD): the new AD object after applying addition
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> y = AD(1, 2, 2, 1)
                >>> x + y
                AD(value: [4], derivatives: [1., 1.])
        """        
        try:
            new_der = self.der + other.der
            new_der2 = self.der2 + other.der2
            new_val = self.val + other.val

            new_tag = np.unique(np.concatenate((self.tag,other.tag),0))#np.nonzero(new_der)
            if self.higher is None or other.higher is None:
                return AD(val = new_val, tag = new_tag, der = new_der, der2 = new_der2, size = self.size)
            else:
                if len(self.higher) != len(other.higher):
                    raise Exception("The two object are not initialized with the same highest order.")
                
                return AD(val = new_val, tag = new_tag, der = new_der, der2 = new_der2, order=len(self.higher),size = self.size,higher=self.higher+other.higher)
        except AttributeError:
            if isinstance(other, int) or isinstance(other, float):

                new_val = self.val + other
                if self.higher is None:
                    new_self = AD(val=new_val, tag=self.tag, der=self.der, der2=self.der2, size=self.size)
                else:
                    new_self = AD(val = new_val, tag = self.tag, der = self.der, der2 = self.der2,order=len(self.higher),size = self.size,higher=self.higher)
                return new_self
            else:
                raise TypeError("Invalid type.")

    def __radd__(self, other):
        """
        Overwrites the __radd__ dunder method to apply addition to an AD object 
        when the AD object is on the right side of the addition sign.
    
                Parameters:
                        self (AD): An AD object to be applied addition to
                        other (AD or int or float): the object to be added to self
    
                Returns:
                        new_self (AD): the new AD object after applying addition
                
                Example:
                >>> x = AD(3, 2, 1, 0)
                >>> 2 + x
                AD(value: [5.], derivatives: [1.])
        """          
        return self + other

    def __iadd__(self, other):
        """
        Overwrites the __iadd__ dunder method to apply addition to an AD object when the operation "+=" is used.
    
                Parameters:
                        self (AD): An AD object to be applied addition to
                        other (AD or int or float): the object to be added to self
    
                Returns:
                        None, but a new AD will be assigned to the original variable.
                
                Example:
                >>> x = AD(3, 2, 1, 0)
                >>> x += 2
                >>> x
                AD(value: [5.], derivatives: [1.])
        """          
        return self + other
    
    # Division
    def __sub__(self, other):
        """
        Overwrites the __sub__ dunder method to apply substraction to an AD object.
    
                Parameters:
                        self (AD): An AD object to be applied substraction to
                        other (AD or valid input for the numpy operation): the object to be substracted from self
    
                Returns:
                        new_self (AD): the new AD object after applying substraction
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> y = AD(1, 2, 2, 1)
                >>> x - y
                AD(value: [2.], derivatives: [1.，-1.])
        """      
        return self + (-1)*other
    
    def __rsub__(self, other):
        """
        Overwrites the __rsub__ dunder method to apply substraction to an AD object.
    
                Parameters:
                        self (AD): An AD object to be applied substraction to
                        other (AD or valid input for the numpy operation): the object to be substracted from self
    
                Returns:
                        new_self (AD): the new AD object after applying substraction
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> 2-x
                AD(value: [-1.], derivatives: [-1.，-0.])
        """      
        
        return (-1)*self + other
    
    def __isub__(self, other):
        """
        Overwrites the __isub__ dunder method to apply substraction to an AD object when the operation "-=" is used.
    
                Parameters:
                        self (AD): An AD object to be applied substraction to
                        other (AD or valid input for the numpy operation): the object to be substracted from self
    
                Returns:
                        None, but a new AD will be assigned to the original variable.

                Example:
                >>> x = AD(3, 2, 1, 0)
                >>> x -= 2
                >>> x 
                AD(value: [1.], derivatives: [1.])
        """          
        return self - other
    
    def __mul__(self, other):
        """
        Overwrites the __mul__ dunder method to apply multiplication to an AD object.
    
                Parameters:
                        self (AD): An AD object to be applied multiplication to
                        other (AD or int or float): the object to be added to self
    
                Returns:
                        new_self (AD): the new AD object after applying multiplication

                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> y = AD(1, 2, 2, 1)
                >>> x * y
                AD(value: [3.], derivatives: [1.，3.])
        """        
        try:
            
            new_der = self.der * other.val + self.val * other.der
            
            new_der2 = self.val * other.der2 + np.matmul(np.array([other.der]).T,np.array([self.der]))  \
                        + np.matmul(np.array([self.der]).T,np.array([other.der]))+ other.val * self.der2

            new_val = self.val * other.val
            
            new_tag = np.unique(np.concatenate((self.tag,other.tag),0))
            # return AD(val = new_val, tag = new_tag, der = new_der, der2 = new_der2, size = self.size)
            higher_der = None
            if self.higher is not None and other.higher is not None and self.tag == other.tag :
                #print("???")
                higher_der = np.array([0.0] * len(self.higher))
                higher_der[0] = new_der
                higher_der[1] = new_der2
                for i in range(0,len(self.higher)):
                    sumval = 0
                    n = i + 1
                    for k in range(n+1):
                        if k == 0:
                            sumval += choose(n,k) * self.val * other.higher[n-k-1]
                        elif k == n:
                            sumval += choose(n,k) * self.higher[k-1] * other.val
                        else:
                            sumval += choose(n,k) * self.higher[k-1] * other.higher[n-k-1]
                    higher_der[i] = sumval
            
            return AD(val=new_val,tag = new_tag, der=new_der, der2=new_der2, order = self.order, size=self.size, higher=higher_der)
        except AttributeError:
            if isinstance(other, int) or isinstance(other, float):
                new_val = self.val * other
                new_der = self.der * other 
                new_der2 = self.der2 * other
                if self.higher is None:
                    new_self = AD(val = new_val, tag = self.tag, der = new_der, der2 = new_der2, size = self.size)
                else:
                    higher_der = other * self.higher
                    new_self = AD(val = new_val, tag = self.tag, der = new_der, der2 = new_der2, order=len(higher_der), size = self.size, higher=higher_der)
                return new_self
                    
            else:
                raise TypeError("Invalid type.")

    def __rmul__(self, other):
        """
        Overwrites the __rmul__ dunder method to apply multiplication to an AD object 
        when the AD object is on the right side of the multiplication sign.
    
                Parameters:
                        self (AD): An AD object to be applied multiplication to
                        other (AD or int or float): the object to be multiplied to self
    
                Returns:
                        new_self (AD): the new AD object after applying multiplication
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> 2*x 
                AD(value: [6.], derivatives: [2.，0.])
        """            
        return self * other

    def __imul__(self, other):
        """
        Overwrites the __imul__ dunder method to apply multiplication to an AD object when the operation "*=" is used.
    
                Parameters:
                        self (AD): An AD object to be applied multiplication to
                        other (AD or int or float): the object to be multiplied to self
    
                Returns:
                        None, but a new AD will be assigned to the original variable.
                
                Example:
                >>> x = AD(3, 2, 1, 0)
                >>> x *= 5
                >>> x
                AD(value: [15.], derivatives: [5.])
        """          
        return self * other
    
    
    ## Division
    def __truediv__(self, other):
        """
        Overwrites the __truediv__ dunder method to apply division to an AD object.
    
                Parameters:
                        self (AD): An AD object to be applied division to
                        other (AD or int or float): the object that self is divided by
    
                Returns:
                        new_self (AD): the new AD object after applying division

                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> y = AD(1, 2, 2, 1)
                >>> x / y
                AD(value: [3.], derivatives: [1.，-2.])
        """
        return self * (other ** (-1.0))

    
    def __rtruediv__(self, other):
        """
        Overwrites the __rtruediv__ dunder method to apply division to an AD object 
        when the AD object is on the right side of the division sign.
    
                Parameters:
                        self (AD): An AD object to be applied division to
                        other (AD or int or float): the object that self is divided by
    
                Returns:
                        new_self (AD): the new AD object after applying division
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> 3/x 
                AD(value: [1.], derivatives: [-0.33333333, -0.])
        """            
        try:
            return other / self
        
        except RecursionError:
            if isinstance(other, int) or isinstance(other, float):
                return other*self **(-1)            
            else:
                raise TypeError("Invalid division type.")

    
    def __itruediv__(self, other):
        """
        Overwrites the __itruediv__ dunder method to apply division to an AD object when the operation "/=" is used.
    
                Parameters:
                        self (AD): An AD object to be applied division to
                        other (AD or int or float): the object that self is divided by
    
                Returns:
                        None, but a new AD will be assigned to the original variable.
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> x /= 3
                >>> x 
                AD(value: [1.], derivatives: [0.33333333, 0.])
        """          
        return self / other
    

    def __pow__(self, other):
        """
        Overwrites the __pow__ dunder method to apply power function to an AD object.
    
                Parameters:
                        self (AD): An AD object to be applied power function to
                        other (AD or int or float): the object that self's power will be raised to
    
                Returns:
                        new_self (AD): the new AD object after applying power function
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> y = AD(1, 2, 2, 1)
                >>> x ** y
                AD(value: [3.], derivatives: [1., 3.29583687])
        """
        if isinstance(other, AD):
            if self.val[0] != 0:
                return exp(log(self)*other)
            else:
                if other.val <= 2:
                    raise ValueError("Derivative is undefined.")
                else:
                    return AD(0, tag = self.tag, der = np.zeros(self.size), der2 = np.zeros((self.size, self.size)))
    
        elif isinstance(other, int) or isinstance(other, float) or isinstance(other, list) or isinstance(other, np.ndarray):
            try:
                other = float(other)
            except:
                other = np.array([float(i) for i in other])

            new_der = (self.val ** (other - 1.)) * other
            new_der2 = (self.val ** (other - 2.)) * other * (other-1.0)

            new_val = np.power(self.val, other)
            higher_der = None
            if self.higher is not None:
                higher_der = np.array([1.0]*len(self.higher))
                for i in range(len(self.higher)):
                    n = i + 1

                    coef = fact_ad(other,n)
                    mainval = math.pow(self.val[0],other-n)
                    higher_der[i] = coef*mainval
            
            return chain_rule(self, new_val, new_der, new_der2, higher_der = higher_der)

        else:
            raise TypeError("Invalid power type.")

    def __ipow__(self, other):
        """
        Overwrites the __ipow__ dunder method to apply power function to an AD object when the operation "**=" is used.
    
                Parameters:
                        self (AD): An AD object to be applied power function to
                        other (AD or int or float): the object that self's power will be raised to
    
                Returns:
                        new_self (AD): the new AD object after applying power function
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> 3 ** x
                AD(value: [27.], derivatives: [29.66253179, 0.])
        """          
        return self ** other

    def __rpow__(self, other):
        """
        Overwrites the __rpow__ dunder method to apply power function to an AD object 
        when the AD object is on the right side of the power sign.
    
                Parameters:
                        self (AD): An AD object to be applied power function to
                        other (AD or int or float): the object that self's power will be raised to
    
                Returns:
                        None, but a new AD will be assigned to the original variable.
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> x **= 2
                >>> x
                AD(value: [9.], derivatives: [6., 0.])
        """            
        try:
            return other ** self
        
        except RecursionError:
            if isinstance(other, int) or isinstance(other, float):
                return exp(log(other) * self)
            else:
                raise TypeError("Invalid type.") 
    

    # Differentiation
    def diff(self, direction=None, order = 1):  
        """
        Calculate and return the derivatives of the function represented by an AD object.
    
                Parameters:
                        self (AD): the AD object whose derivatives will be calculated.
                        direction (int): the seed indicating which variable's derivative should be returned
    
                Returns:
                        derivative (float): the derivative of the AD object, wrt the input direction 
                
                Example:
                >>> x = AD(3, 2, 2, 0)
                >>> y = AD(1, 2, 2, 1)
                >>> f = x / y
                >>> dfdy = f.diff(1)
                >>> dfdy 
                -3.0
                
        """
        if order == 1 and isinstance(direction, int):
            return self.der[direction]

        elif order == 2 and isinstance(direction, list) and len(direction) == 2:
            return self.der2[direction[0], direction[1]]
        else:
            raise Exception("Order exceeds 2 or length of direction and order don't match.")


    # Calculate higher order derivatives
    def higherdiff(self,order):
        """
        Return the derivative of the desired order. Only works for one scalar variable and one scalar function.
    
                Parameters:
                        self (AD): the AD object whose derivatives will be calculated.
                        order (string): the order of derivative
    
                Returns:
                        the derivative(float) of the given order evaluated at the point of self.val
                
                Example:
                >>> x = AD(val = 3, order = 10, size = 1, tag = 0)
                >>> f = x**5
                >>> f.higherdiff(5)
                120.0
                >>> f.higherdiff(6)
                0.0

        """
        if not isinstance(order, numbers.Integral):
            raise TypeError("Highest order of derivatives must be a positive integer.")
        elif order < 1:
            raise ValueError("Highest order of derivatives must be at least 1.")
        elif self.higher is None:
            raise Exception("You didn't initialize higher order")
        elif order > len(self.higher):
            raise ValueError("You asked for an order beyond what you stored.")

        return self.higher[order-1]






# -*- coding: utf-8 -*-
"""
Functionality for working with a git repository
"""

from os.path import isdir
from os.path import join as join_path

from . import shell, utils
from .exceptions import ShellCommandException


def git_tree_add(struct, prefix="", **kwargs):
    """Adds recursively a directory structure to git

    Args:
        struct (dict): directory structure as dictionary of dictionaries
        prefix (str): prefix for the given directory structure

    Additional keyword arguments are passed to the
    :obj:`git <pyscaffold.shell.ShellCommand>` callable object.
    """
    for name, content in struct.items():
        if isinstance(content, str):
            shell.git('add', join_path(prefix, name), **kwargs)
        elif isinstance(content, dict):
            git_tree_add(
                struct[name], prefix=join_path(prefix, name), **kwargs)
        elif content is None:
            shell.git('add', join_path(prefix, name), **kwargs)
        else:
            raise RuntimeError("Don't know what to do with content type "
                               "{type}.".format(type=type(content)))


def add_tag(project, tag_name, message=None, **kwargs):
    """Add an (annotated) tag to the git repository.

    Args:
        project (str): path to the project
        tag_name (str): name of the tag
        message (str): optional tag message

    Additional keyword arguments are passed to the
    :obj:`git <pyscaffold.shell.ShellCommand>` callable object.
    """
    with utils.chdir(project):
        if message is None:
            shell.git('tag', tag_name, **kwargs)
        else:
            shell.git('tag', '-a', tag_name, '-m', message, **kwargs)


def init_commit_repo(project, struct, **kwargs):
    """Initialize a git repository

    Args:
        project (str): path to the project
        struct (dict): directory structure as dictionary of dictionaries

    Additional keyword arguments are passed to the
    :obj:`git <pyscaffold.shell.ShellCommand>` callable object.
    """
    with utils.chdir(project, pretend=kwargs.get('pretend')):
        shell.git('init', **kwargs)
        git_tree_add(struct[project], **kwargs)
        shell.git('commit', '-m', 'Initial commit', **kwargs)


def is_git_repo(folder):
    """Check if a folder is a git repository

    Args:
        folder (str): path
    """
    if not isdir(folder):
        return False

    with utils.chdir(folder):
        try:
            shell.git('rev-parse', '--git-dir')
        except ShellCommandException:
            return False
        return True


def get_git_root(default=None):
    """Return the path to the top-level of the git repository or *default*.

    Args:
        default (str): if no git root is found, default is returned

    Returns:
        str: top-level path or *default*
    """
    if shell.git is None:
        return default
    try:
        return next(shell.git('rev-parse', '--show-toplevel'))
    except ShellCommandException:
        return default



# -*- coding: utf-8 -*-
"""
Extension that generates configuration files for Yelp `pre-commit`_.

.. _pre-commit: http://pre-commit.com
"""

from ..api import Extension, helpers
from ..log import logger
from ..templates import isort_cfg, pre_commit_config


class PreCommit(Extension):
    """Generate pre-commit configuration file"""
    def activate(self, actions):
        """Activate extension

        Args:
            actions (list): list of actions to perform

        Returns:
            list: updated list of actions
        """
        return (
            self.register(actions, self.add_files, after='define_structure') +
            [self.instruct_user])

    @staticmethod
    def add_files(struct, opts):
        """Add .pre-commit-config.yaml file to structure

        Since the default template uses isort, this function also provides an
        initial version of .isort.cfg that can be extended by the user
        (it contains some useful skips, e.g. tox and venv)

        Args:
            struct (dict): project representation as (possibly) nested
                :obj:`dict`.
            opts (dict): given options, see :obj:`create_project` for
                an extensive list.

        Returns:
            struct, opts: updated project representation and options
        """
        files = {
            '.pre-commit-config.yaml': (
                pre_commit_config(opts), helpers.NO_OVERWRITE
            ),
            '.isort.cfg': (
                isort_cfg(opts), helpers.NO_OVERWRITE
            ),
        }

        return helpers.merge(struct, {opts['project']: files}), opts

    @staticmethod
    def instruct_user(struct, opts):
        logger.warning(
            '\nA `.pre-commit-config.yaml` file was generated inside your '
            'project but in order to make sure the hooks will run, please '
            'don\'t forget to install the `pre-commit` package:\n\n'
            '  cd %s\n'
            '  # it is a good idea to create and activate a virtualenv here\n'
            '  pip install pre-commit\n'
            '  pre-commit install\n'
            '  # another good idea is update the hooks to the latest version\n'
            '  # pre-commit autoupdate\n\n'
            'You might also consider including similar instructions in your '
            'docs, to remind the contributors to do the same.\n',
            opts['project'])

        return struct, opts



# -*- coding: utf-8 -*-
"""
Shell commands like git, django-admin.py etc.
"""

import functools
import shutil
import subprocess
import sys

from .exceptions import ShellCommandException
from .log import logger


class ShellCommand(object):
    """Shell command that can be called with flags like git('add', 'file')

    Args:
        command (str): command to handle
        shell (bool): run the command in the shell
        cwd (str): current working dir to run the command

    The produced command can be called with the following keyword arguments:

        - **log** (*bool*): log activity when true. ``False`` by default.
        - **pretend** (*bool*): skip execution (but log) when pretending.
          ``False`` by default.

    The positional arguments are passed to the underlying shell command.
    """
    def __init__(self, command, shell=True, cwd=None):
        self._command = command
        self._shell = shell
        self._cwd = cwd

    def __call__(self, *args, **kwargs):
        """Execute command with the given arguments."""
        command = "{cmd} {args}".format(cmd=self._command,
                                        args=subprocess.list2cmdline(args))

        should_pretend = kwargs.get('pretend')
        should_log = kwargs.get('log', should_pretend)
        # ^ When pretending, automatically output logs
        #   (after all, this is the primary purpose of pretending)

        if should_log:
            logger.report('run', command, context=self._cwd)

        if should_pretend:
            output = ''
        else:
            try:
                output = subprocess.check_output(command,
                                                 shell=self._shell,
                                                 cwd=self._cwd,
                                                 stderr=subprocess.STDOUT,
                                                 universal_newlines=True)
            except subprocess.CalledProcessError as e:
                raise ShellCommandException(e.output) from e

        return (line for line in output.splitlines())


def shell_command_error2exit_decorator(func):
    """Decorator to convert given ShellCommandException to an exit message

    This avoids displaying nasty stack traces to end-users
    """
    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except ShellCommandException as e:
            e = e.__cause__
            print("{err}:\n{msg}".format(err=str(e), msg=e.output))
            sys.exit(1)
    return func_wrapper


def get_git_cmd(**args):
    """Retrieve the git shell command depending on the current platform

    Args:
        **args: additional keyword arguments to :obj:`~.ShellCommand`
    """
    if sys.platform == "win32":
        for cmd in ["git.cmd", "git.exe"]:
            git = ShellCommand(cmd, **args)
            try:
                git("--version")
            except ShellCommandException:
                continue
            return git
        else:
            return None
    else:
        git = ShellCommand("git", **args)
        try:
            git("--version")
        except ShellCommandException:
            return None
        return git


def command_exists(cmd):
    """Check if command exists

    Args:
        cmd: executable name
    """
    if shutil.which(cmd) is None:
        return False
    else:
        return True


#: Command for git
git = get_git_cmd()

#: Command for django-admin.py
django_admin = ShellCommand("django-admin.py")



# -*- coding: utf-8 -*-
"""
Basic support for ANSI code formatting.
"""

import sys

ESCAPE = '\033[{:d}m'
STYLES = dict(
    clear=0,
    bold=1,
    black=30,
    red=31,
    green=32,
    yellow=33,
    blue=34,
    magenta=35,
    cyan=36,
    white=37,
    on_black=40,
    on_red=41,
    on_green=42,
    on_yellow=43,
    on_blue=44,
    on_magenta=45,
    on_cyan=46,
    on_white=47
)


def isatty(stream=None):
    """Detect if the given stream/stdout is part of an interactive terminal.

    Args:
        stream: optionally the stream to check

    Returns:
        bool: result of check
    """
    stream = stream or sys.stdout

    if hasattr(stream, 'isatty'):
        return stream.isatty()

    return False


def init_colorama():
    """Initialize colorama if it is available.

    Returns:
        bool: result of check
    """
    try:
        import colorama  # noqa
        colorama.init()
        return True
    except ImportError:
        return False


def curses_available():
    """Check if the curses package from stdlib is available.

    Usually not available for windows, but its presence indicates that the
    terminal is capable of displaying some UI.

    Returns:
        bool: result of check
    """
    try:
        import curses  # noqa
        return True
    except ImportError:
        return False


SYSTEM_SUPPORTS_COLOR = curses_available() or init_colorama()
# Eagerly executed, in order to avoid calling colorama.init multiple times


def supports_color(stream=None):
    """Check if the stream is supposed to handle coloring.

    Returns:
        bool: result of check
    """
    return isatty(stream) and SYSTEM_SUPPORTS_COLOR


def decorate(msg, *styles):
    """Use ANSI codes to format the message.

    Args:
        msg (str): string to be formatted
        *styles (list): the remaining arguments should be strings that
            represent the 8 basic ANSI colors. ``clear`` and ``bold`` are also
            supported. For background colors use ``on_<color>``.

    Returns:
        str: styled and formatted message
    """
    if not styles:
        return msg

    styles = ''.join(ESCAPE.format(STYLES[s]) for s in styles if s in STYLES)
    return styles + msg + ESCAPE.format(STYLES['clear'])



# -*- coding: utf-8 -*-
"""
Miscellaneous utilities and tools
"""

import errno
import functools
import keyword
import logging
import os
import re
import shutil
import sys
import traceback
from contextlib import contextmanager
from pathlib import Path

from pkg_resources import parse_version

from . import __version__ as pyscaffold_version
from .contrib.setuptools_scm.version import VERSION_CLASS
from .exceptions import InvalidIdentifier, OldSetuptools
from .log import logger


@contextmanager
def _chdir_logging_context(path, should_log):
    """Private auxiliar function for logging inside chdir"""
    if should_log:
        logger.report('chdir', path)
        with logger.indent():
            yield
    else:
        yield


@contextmanager
def chdir(path, **kwargs):
    """Contextmanager to change into a directory

    Args:
        path (str): path to change current working directory to

    Keyword Args:
        log (bool): log activity when true. Default: ``False``.
        pretend (bool): skip execution (but log) when pretending.
            Default ``False``.
    """
    should_pretend = kwargs.get('pretend')
    should_log = kwargs.get('log', should_pretend)
    # ^ When pretending, automatically output logs
    #   (after all, this is the primary purpose of pretending)

    curr_dir = os.getcwd()

    try:
        with _chdir_logging_context(path, should_log):
            if not should_pretend:
                # ToDo: Remove str when we require PY 3.6
                os.chdir(str(path))  # str to handle pathlib args
            yield
    finally:
        os.chdir(curr_dir)


def move(*src, **kwargs):
    """Move files or directories to (into) a new location

    Args:
        *src (str[]): one or more files/directories to be moved

    Keyword Args:
        target (str): if target is a directory, ``src`` will be moved inside
            it. Otherwise, it will be the new path (note that it may be
            overwritten)
        log (bool): log activity when true. Default: ``False``.
        pretend (bool): skip execution (but log) when pretending.
            Default ``False``.
    """
    target = kwargs['target']  # Required arg
    should_pretend = kwargs.get('pretend')
    should_log = kwargs.get('log', should_pretend)
    # ^ When pretending, automatically output logs
    #   (after all, this is the primary purpose of pretending)

    for path in src:
        if not should_pretend:
            shutil.move(path, target)
        if should_log:
            logger.report('move', path, target=target)


def is_valid_identifier(string):
    """Check if string is a valid package name

    Args:
        string (str): package name

    Returns:
        bool: True if string is valid package name else False
    """
    if not re.match("[_A-Za-z][_a-zA-Z0-9]*$", string):
        return False
    if keyword.iskeyword(string):
        return False
    return True


def make_valid_identifier(string):
    """Try to make a valid package name identifier from a string

    Args:
        string (str): invalid package name

    Returns:
        str: valid package name as string or :obj:`RuntimeError`

    Raises:
        :obj:`InvalidIdentifier`: raised if identifier can not be converted
    """
    string = string.strip()
    string = string.replace("-", "_")
    string = string.replace(" ", "_")
    string = re.sub('[^_a-zA-Z0-9]', '', string)
    string = string.lower()
    if is_valid_identifier(string):
        return string
    else:
        raise InvalidIdentifier(
                "String cannot be converted to a valid identifier.")


def exceptions2exit(exception_list):
    """Decorator to convert given exceptions to exit messages

    This avoids displaying nasty stack traces to end-users

    Args:
        exception_list [Exception]: list of exceptions to convert
    """
    def exceptions2exit_decorator(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except tuple(exception_list) as e:
                if logger.level <= logging.DEBUG:
                    # user surely wants to see the stacktrace
                    traceback.print_exc()
                print("ERROR: {}".format(e))
                sys.exit(1)
        return func_wrapper
    return exceptions2exit_decorator


# from http://en.wikibooks.org/, Creative Commons Attribution-ShareAlike 3.0
def levenshtein(s1, s2):
    """Calculate the Levenshtein distance between two strings

    Args:
        s1 (str): first string
        s2 (str): second string

    Returns:
        int: distance between s1 and s2
    """
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def prepare_namespace(namespace_str):
    """Check the validity of namespace_str and split it up into a list

    Args:
        namespace_str (str): namespace, e.g. "com.blue_yonder"

    Returns:
        [str]: list of namespaces, e.g. ["com", "com.blue_yonder"]

    Raises:
          :obj:`InvalidIdentifier` : raised if namespace is not valid
    """
    namespaces = namespace_str.split('.') if namespace_str else list()
    for namespace in namespaces:
        if not is_valid_identifier(namespace):
            raise InvalidIdentifier(
                "{} is not a valid namespace package.".format(namespace))
    return ['.'.join(namespaces[:i+1]) for i in range(len(namespaces))]


def check_setuptools_version():
    """Check minimum required version of setuptools

    Check that setuptools has all necessary capabilities for setuptools_scm
    as well as support for configuration with the help of ``setup.cfg``.

    Raises:
          :obj:`OldSetuptools` : raised if necessary capabilities are not met
    """
    try:
        from setuptools import __version__ as setuptools_ver
        from pkg_resources import parse_version
    except ImportError:
        raise OldSetuptools

    setuptools_too_old = parse_version(setuptools_ver) < parse_version('38.3')
    setuptools_scm_check_failed = VERSION_CLASS is None
    if setuptools_too_old or setuptools_scm_check_failed:
        raise OldSetuptools


def create_file(path, content, pretend=False):
    """Create a file in the given path.

    This function reports the operation in the logs.

    Args:
        path (str): path in the file system where contents will be written.
        content (str): what will be written.
        pretend (bool): false by default. File is not written when pretending,
            but operation is logged.
    """
    if not pretend:
        with open(path, 'w', encoding='utf-8') as fh:
            fh.write(content)

    logger.report('create', path)


def create_directory(path, update=False, pretend=False):
    """Create a directory in the given path.

    This function reports the operation in the logs.

    Args:
        path (str): path in the file system where contents will be written.
        update (bool): false by default. A :obj:`OSError` is raised when update
            is false and the directory already exists.
        pretend (bool): false by default. Directory is not created when
            pretending, but operation is logged.
    """
    if not pretend:
        try:
            os.mkdir(path)
        except OSError:
            if not update:
                raise
            return  # Do not log if not created

    logger.report('create', path)


def dasherize(word):
    """Replace underscores with dashes in the string.

    Example::

        >>> dasherize("foo_bar")
        "foo-bar"

    Args:
        word (str): input word

    Returns:
        input word with underscores replaced by dashes
    """
    return word.replace('_', '-')


def get_id(function):
    """Given a function, calculate its identifier.

    A identifier is a string in the format ``<module name>:<function name>``,
    similarly to the convention used for setuptools entry points.

    Note:
        This function does not return a Python 3 ``__qualname__`` equivalent.
        If the function is nested inside another function or class, the parent
        name is ignored.

    Args:
        function (callable): function object

    Returns:
        str: identifier
    """
    return '{}:{}'.format(function.__module__, function.__name__)


def get_setup_requires_version():
    """Determines the proper `setup_requires` string for PyScaffold

    E.g. setup_requires = pyscaffold>=3.0a0,<3.1a0

    Returns:
        str: requirement string for setup_requires
    """
    require_str = "pyscaffold>={major}.{minor}a0,<{major}.{next_minor}a0"
    major, minor, *rest = (parse_version(pyscaffold_version)
                           .base_version.split('.'))
    next_minor = int(minor) + 1
    return require_str.format(major=major, minor=minor, next_minor=next_minor)


def localize_path(path_string):
    """Localize path for Windows, Unix, i.e. / or \

    Args:
        path_string (str): path using /

    Returns:
        str: path depending on OS
    """
    return str(Path(path_string))


#: Windows-specific error code indicating an invalid pathname.
ERROR_INVALID_NAME = 123


def is_pathname_valid(pathname):
    """Check if a pathname is valid

    Code by Cecil Curry from StackOverflow

    Args:
        pathname (str): string to validate

    Returns:
        `True` if the passed pathname is a valid pathname for the current OS;
        `False` otherwise.
    """
    # If this pathname is either not a string or is but is empty, this pathname
    # is invalid.
    try:
        if not isinstance(pathname, str) or not pathname:
            return False

        # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
        # if any. Since Windows prohibits path components from containing `:`
        # characters, failing to strip this `:`-suffixed prefix would
        # erroneously invalidate all valid absolute Windows pathnames.
        _, pathname = os.path.splitdrive(pathname)

        # Directory guaranteed to exist. If the current OS is Windows, this is
        # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
        # environment variable); else, the typical root directory.
        root_dirname = os.environ.get('HOMEDRIVE', 'C:') \
            if sys.platform == 'win32' else os.path.sep
        assert os.path.isdir(root_dirname)  # ...Murphy and her ironclad Law

        # Append a path separator to this directory if needed.
        root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

        # Test whether each path component split from this pathname is valid or
        # not, ignoring non-existent and non-readable path components.
        for pathname_part in pathname.split(os.path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            # If an OS-specific exception is raised, its error code
            # indicates whether this pathname is valid or not. Unless this
            # is the case, this exception implies an ignorable kernel or
            # filesystem complaint (e.g., path not found or inaccessible).
            #
            # Only the following exceptions indicate invalid pathnames:
            #
            # * Instances of the Windows-specific "WindowsError" class
            #   defining the "winerror" attribute whose value is
            #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
            #   fine-grained and hence useful than the generic "errno"
            #   attribute. When a too-long pathname is passed, for example,
            #   "errno" is "ENOENT" (i.e., no such file or directory) rather
            #   than "ENAMETOOLONG" (i.e., file name too long).
            # * Instances of the cross-platform "OSError" class defining the
            #   generic "errno" attribute whose value is either:
            #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
            #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
            except OSError as exc:
                if hasattr(exc, 'winerror'):
                    if exc.winerror == ERROR_INVALID_NAME:
                        return False
                elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                    return False
    # If a "TypeError" exception was raised, it almost certainly has the
    # error message "embedded NUL character" indicating an invalid pathname.
    except TypeError:
        return False
    # If no exception was raised, all path components and hence this
    # pathname itself are valid. (Praise be to the curmudgeonly python.)
    else:
        return True
    # If any other exception was raised, this is an unrelated fatal issue
    # (e.g., a bug). Permit this exception to unwind the call stack.
    #
    # Did we mention this should be shipped with Python already?


def on_ro_error(func, path, exc_info):
    """Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``

    Args:
        func (callable): function which raised the exception
        path (str): path passed to `func`
        exc_info (tuple of str): exception info returned by sys.exc_info()
    """
    import stat
    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


def rm_rf(path):
    """Remove a path by all means like `rm -rf` in Linux.

    Args (str): Path to remove:
    """
    shutil.rmtree(path, onerror=on_ro_error)



# -*- coding: utf-8 -*-
"""
Custom exceptions used by PyScaffold to identify common deviations from the
expected behavior.
"""


class ActionNotFound(KeyError):
    """Impossible to find the required action."""

    def __init__(self, name, *args, **kwargs):
        message = ActionNotFound.__doc__[:-1] + ': `{}`'.format(name)
        super().__init__(message, *args, **kwargs)


class DirectoryAlreadyExists(RuntimeError):
    """The project directory already exists, but no ``update`` or ``force``
    option was used.
    """


class DirectoryDoesNotExist(RuntimeError):
    """No directory was found to be updated."""


class GitNotInstalled(RuntimeError):
    """PyScaffold requires git to run."""

    DEFAULT_MESSAGE = "Make sure git is installed and working."

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class GitNotConfigured(RuntimeError):
    """PyScaffold tries to read user.name and user.email from git config."""

    DEFAULT_MESSAGE = (
        'Make sure git is configured. Run:\n'
        '  git config --global user.email "you@example.com"\n'
        '  git config --global user.name "Your Name"\n'
        "to set your account's default identity.")

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class GitDirtyWorkspace(RuntimeError):
    """Workspace of git is empty."""

    DEFAULT_MESSAGE = (
        "Your working tree is dirty. Commit your changes first"
        " or use '--force'.")

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class InvalidIdentifier(RuntimeError):
    """Python requires a specific format for its identifiers.

    https://docs.python.org/3.6/reference/lexical_analysis.html#identifiers
    """


class OldSetuptools(RuntimeError):
    """PyScaffold requires a recent version of setuptools."""

    DEFAULT_MESSAGE = (
        "Your setuptools version is too old (<38.3). "
        "Use `pip install -U setuptools` to upgrade.\n"
        "If you have the deprecated `distribute` package installed "
        "remove it or update to version 0.7.3.")

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class PyScaffoldTooOld(RuntimeError):
    """PyScaffold cannot update a pre 3.0 version"""

    DEFAULT_MESSAGE = (
        "setup.cfg has no section [pyscaffold]! "
        "Are you trying to update a pre 3.0 version?")

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class NoPyScaffoldProject(RuntimeError):
    """PyScaffold cannot update a project that it hasn't generated"""

    DEFAULT_MESSAGE = (
        "Could not update project. Was it generated with PyScaffold?")

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class ShellCommandException(RuntimeError):
    """Outputs proper logging when a ShellCommand fails"""

    def __init__(self, message, *args, **kwargs):
        super().__init__(message, *args, **kwargs)



# -*- coding: utf-8 -*-
"""
Contribution packages used by PyScaffold

All packages inside ``contrib`` are external packages that come with their
own licences and are not part of the PyScaffold source code itself.
The reason for shipping these dependencies directly is to avoid problems in
the resolution of ``setup_requires`` dependencies that occurred more often 
than not, see issues #71 and #72.

Currently the contrib packages are:

1) setuptools_scm v3.3.3
2) pytest-runner 5.1
3) configupdater 1.0

The packages/modules were just copied over.
"""

# Following dummy definitions are here in case PyScaffold version < 3
# is still installed and setuptools checks the registered entry_points.
SCM_HG_FILES_COMMAND = ''
SCM_GIT_FILES_COMMAND = ''


def warn_about_deprecated_pyscaffold():
    raise RuntimeError("A PyScaffold version less than 3.0 was detected, "
                       "please upgrade!")


def scm_find_files(*args, **kwargs):
    warn_about_deprecated_pyscaffold()


def scm_parse_hg(*args, **kwargs):
    warn_about_deprecated_pyscaffold()


def scm_parse_git(*args, **kwargs):
    warn_about_deprecated_pyscaffold()


def scm_parse_archival(*args, **kwargs):
    warn_about_deprecated_pyscaffold()


def scm_parse_pkginfo(*args, **kwargs):
    warn_about_deprecated_pyscaffold()


def scm_guess_next_dev_version(*args, **kwargs):
    warn_about_deprecated_pyscaffold()


def scm_postrelease_version(*args, **kwargs):
    warn_about_deprecated_pyscaffold()


def scm_get_local_node_and_date(*args, **kwargs):
    warn_about_deprecated_pyscaffold()


def scm_get_local_dirty_tag(*args, **kwargs):
    warn_about_deprecated_pyscaffold()


def write_pbr_json(*args, **kwargs):
    warn_about_deprecated_pyscaffold()



# -*- coding: utf-8 -*-
"""
Custom logging infrastructure to provide execution information for the user.
"""

from collections import defaultdict
from contextlib import contextmanager
from logging import INFO, Formatter, LoggerAdapter, StreamHandler, getLogger
from os.path import realpath, relpath
from os.path import sep as pathsep

from . import termui

DEFAULT_LOGGER = __name__


def _are_equal_paths(path1, path2):
    return realpath(path1) == realpath(path2)


def _is_current_path(path):
    return _are_equal_paths(path, '.')


def configure_logger(opts):
    """Configure the default logger

    Args:
        opts (dict): command line parameters
    """
    if 'log_level' in opts:
        logger.level = opts['log_level']

    # if terminal supports, use colors
    stream = getattr(logger.handler, 'stream', None)
    if termui.supports_color(stream):
        logger.formatter = ColoredReportFormatter()
        logger.handler.setFormatter(logger.formatter)


class ReportFormatter(Formatter):
    """Formatter that understands custom fields in the log record."""

    ACTIVITY_MAXLEN = 12
    SPACING = '  '
    CONTEXT_PREFIX = 'from'
    TARGET_PREFIX = 'to'

    def format(self, record):
        """Compose message when a record with report information is given."""
        if hasattr(record, 'activity'):
            return self.format_report(record)

        return self.format_default(record)

    def create_padding(self, activity):
        """Create the appropriate padding in order to align activities."""
        actual = len(activity)
        count = max(self.ACTIVITY_MAXLEN - actual, 0)
        return ' ' * count

    def format_path(self, path):
        """Simplify paths to avoid wasting space in terminal."""
        from .utils import is_pathname_valid  # late import due to cycles

        if is_pathname_valid(path) and pathsep in path:
            # Heuristic to determine if subject is a file path
            # that needs to be made short
            abbrev = relpath(path)

            if len(abbrev) < len(path):
                # Ignore if not shorter
                path = abbrev

        return path

    def format_activity(self, activity):
        """Format the activity keyword."""
        return activity

    # Subclasses may need the activity name to choose correct format,
    # so the following 3 methods accept a second parameter
    # (even if they not use it)
    def format_subject(self, subject, _activity=None):
        """Format the subject of the activity."""
        return self.format_path(subject)

    def format_target(self, target, _activity=None):
        """Format extra information about the activity target."""
        if target and not _is_current_path(target):
            return "{} '{}'".format(self.TARGET_PREFIX,
                                    self.format_path(target))

        return ''

    def format_context(self, context, _activity=None):
        """Format extra information about the activity context."""
        if context and not _is_current_path(context):
            return "{} '{}'".format(self.CONTEXT_PREFIX,
                                    self.format_path(context))

        return ''

    def format_default(self, record):
        """Format default log messages."""
        record.msg = self.SPACING * max(record.nesting, 0) + record.msg

        return super(ReportFormatter, self).format(record)

    def format_report(self, record):
        """Compose message when a custom record is given."""
        activity = record.activity
        record.msg = (
            self.create_padding(activity) +
            self.format_activity(activity) +
            self.SPACING * max(record.nesting + 1, 0) +
            ' '.join([
                text for text in [
                    self.format_subject(record.subject, activity),
                    self.format_target(record.target, activity),
                    self.format_context(record.context, activity)
                ] if text  # Filter empty strings
            ])
        )

        return super(ReportFormatter, self).format(record)


class ColoredReportFormatter(ReportFormatter):
    """Format logs with ANSI colors."""

    ACTIVITY_STYLES = defaultdict(
        lambda: ('blue', 'bold'),
        create=('green', 'bold'),
        move=('green', 'bold'),
        remove=('red', 'bold'),
        delete=('red', 'bold'),
        skip=('yellow', 'bold'),
        run=('magenta', 'bold'),
        invoke=('bold',)
    )

    SUBJECT_STYLES = defaultdict(
        tuple,
        invoke=('blue',)
    )

    LOG_STYLES = defaultdict(
        tuple,
        debug=('green',),
        info=('blue',),
        warning=('yellow',),
        error=('red',),
        critical=('red', 'bold')
    )

    CONTEXT_PREFIX = termui.decorate(
        ReportFormatter.CONTEXT_PREFIX, 'magenta', 'bold')

    TARGET_PREFIX = termui.decorate(
        ReportFormatter.TARGET_PREFIX, 'magenta', 'bold')

    def format_activity(self, activity):
        return termui.decorate(activity, *self.ACTIVITY_STYLES[activity])

    def format_subject(self, subject, activity=None):
        parent = super(ColoredReportFormatter, self)
        subject = parent.format_subject(subject, activity)
        return termui.decorate(subject, *self.SUBJECT_STYLES[activity])

    def format_default(self, record):
        record.msg = termui.decorate(
            record.msg, *self.LOG_STYLES[record.levelname.lower()])
        return super(ColoredReportFormatter, self).format_default(record)


class ReportLogger(LoggerAdapter):
    """Suitable wrapper for PyScaffold CLI interactive execution reports.

    Args:
        logger (logging.Logger): custom logger to be used. Optional: the
            default logger will be used.
        handlers (logging.Handler): custom logging handler to be used.
            Optional: a :class:`logging.StreamHandler` is used by default.
        formatter (logging.Formatter): custom formatter to be used.
            Optional: by default a :class:`~.ReportFormatter` is created and
            used.
        extra (dict): extra attributes to be merged into the log record.
            Options, empty by default.

    Attributes:
        wrapped (logging.Logger): underlying logger object.
        handler (logging.Handler): stream handler configured for
            providing user feedback in PyScaffold CLI.
        formatter (logging.Formatter): formatter configured in the
            default handler.
        nesting (int): current nesting level of the report.
    """

    def __init__(self, logger=None, handler=None, formatter=None, extra=None):
        self.nesting = 0
        self.wrapped = logger or getLogger(DEFAULT_LOGGER)
        self.extra = extra or {}
        self.handler = handler or StreamHandler()
        self.formatter = formatter or ReportFormatter()
        self.handler.setFormatter(self.formatter)
        self.wrapped.addHandler(self.handler)
        super(ReportLogger, self).__init__(self.wrapped, self.extra)

    @property
    def level(self):
        """Effective level of the logger"""
        return self.wrapped.getEffectiveLevel()

    @level.setter
    def level(self, value):
        """Set the logger level"""
        self.wrapped.setLevel(value)

    def process(self, msg, kwargs):
        """Method overridden to augment LogRecord with the `nesting` attribute.
        """
        (msg, kwargs) = super(ReportLogger, self).process(msg, kwargs)
        extra = kwargs.get('extra', {})
        extra['nesting'] = self.nesting
        kwargs['extra'] = extra
        return msg, kwargs

    def report(self, activity, subject,
               context=None, target=None, nesting=None, level=INFO):
        """Log that an activity has occurred during scaffold.

        Args:
            activity (str): usually a verb or command, e.g. ``create``,
                ``invoke``, ``run``, ``chdir``...
            subject (str): usually a path in the file system or an action
                identifier.
            context (str): path where the activity take place.
            target (str): path affected by the activity
            nesting (int): optional nesting level. By default it is calculated
                from the activity name.
            level (int): log level. Defaults to :obj:`logging.INFO`.
                See :mod:`logging` for more information.

        Notes:
            This method creates a custom log record, with additional fields:
            **activity**, **subject**, **context**, **target** and **nesting**,
            but an empty **msg** field. The :class:`ReportFormatter`
            creates the log message from the other fields.

            Often **target** and **context** complement the logs when
            **subject** does not hold all the necessary information. For
            example::

                logger.report('copy', 'my/file', target='my/awesome/path')
                logger.report('run', 'command', context='current/working/dir')
        """
        return self.wrapped.log(level, '', extra={
            'activity': activity,
            'subject': subject,
            'context': context,
            'target': target,
            'nesting': nesting or self.nesting
        })

    @contextmanager
    def indent(self, count=1):
        """Temporarily adjust padding while executing a context.

        Example:

            .. code-block:: python

                from pyscaffold.log import logger
                logger.report('invoke', 'custom_action')
                with logger.indent():
                   logger.report('create', 'some/file/path')

                # Expected logs:
                # --------------------------------------
                #       invoke  custom_action
                #       create    some/file/path
                # --------------------------------------
                # Note how the spacing between activity and subject in the
                # second entry is greater than the equivalent in the first one.

        Note:
            This method is not thread-safe and should be used with care.
        """
        prev = self.nesting
        self.nesting += count
        try:
            yield
        finally:
            self.nesting = prev

    def copy(self):
        """Produce a copy of the wrapped logger.

        Sometimes, it is better to make a copy of th report logger to keep
        indentation consistent.
        """
        clone = self.__class__(self.wrapped, self.handler,
                               self.formatter, self.extra)
        clone.nesting = self.nesting

        return clone


logger = ReportLogger()
"""Default logger configured for PyScaffold."""



# -*- coding: utf-8 -*-
"""
Provide general information about the system, user etc.
"""

import copy
import getpass
import os
import socket
from enum import Enum
from operator import itemgetter

from . import shell
from .exceptions import (
    GitNotConfigured,
    GitNotInstalled,
    PyScaffoldTooOld,
    ShellCommandException
)
from .templates import licenses
from .update import read_setupcfg
from .utils import chdir, levenshtein


class GitEnv(Enum):
    author_name = "GIT_AUTHOR_NAME"
    author_email = "GIT_AUTHOR_EMAIL"
    author_date = "GIT_AUTHOR_DATE"
    committer_name = "GIT_COMMITTER_NAME"
    committer_email = "GIT_COMMITTER_EMAIL"
    committer_date = "GIT_COMMITTER_DATE"


def username():
    """Retrieve the user's name

    Returns:
        str: user's name
    """
    user = os.getenv(GitEnv.author_name.value)
    if user is None:
        try:
            user = next(shell.git("config", "--get", "user.name"))
            user = user.strip()
        except ShellCommandException:
            user = getpass.getuser()
    return user


def email():
    """Retrieve the user's email

    Returns:
        str: user's email
    """
    mail = os.getenv(GitEnv.author_email.value)
    if mail is None:
        try:
            mail = next(shell.git("config", "--get", "user.email"))
            mail = mail.strip()
        except ShellCommandException:
            user = getpass.getuser()
            host = socket.gethostname()
            mail = "{user}@{host}".format(user=user, host=host)
    return mail


def is_git_installed():
    """Check if git is installed

    Returns:
        bool: True if git is installed, False otherwise
    """
    if shell.git is None:
        return False
    try:
        shell.git("--version")
    except ShellCommandException:
        return False
    return True


def is_git_configured():
    """Check if user.name and user.email is set globally in git

    Check first git environment variables, then config settings.
    This will also return false if git is not available at all.

    Returns:
        bool: True if it is set globally, False otherwise
    """
    if (os.getenv(GitEnv.author_name.value) and
            os.getenv(GitEnv.author_email.value)):
        return True
    else:
        try:
            for attr in ("name", "email"):
                shell.git("config", "--get", "user.{}".format(attr))
        except ShellCommandException:
            return False
        else:
            return True


def check_git():
    """Checks for git and raises appropriate exception if not

     Raises:
        :class:`~.GitNotInstalled`: when git command is not available
        :class:`~.GitNotConfigured`: when git does not know user information
    """
    if not is_git_installed():
        raise GitNotInstalled
    if not is_git_configured():
        raise GitNotConfigured


def is_git_workspace_clean(path):
    """Checks if git workspace is clean

    Args:
        path (str): path to git repository

    Returns:
        bool: condition if workspace is clean or not

     Raises:
        :class:`~.GitNotInstalled`: when git command is not available
        :class:`~.GitNotConfigured`: when git does not know user information
    """
    # ToDo: Change to pathlib for v4
    check_git()
    try:
        with chdir(path):
            shell.git('diff-index', '--quiet', 'HEAD', '--')
    except ShellCommandException:
        return False
    return True


def project(opts):
    """Update user options with the options of an existing PyScaffold project

    Params:
        opts (dict): options of the project

    Returns:
        dict: options with updated values

    Raises:
        :class:`~.PyScaffoldTooOld`: when PyScaffold is to old to update from
        :class:`~.NoPyScaffoldProject`: when project was not generated with
            PyScaffold
    """
    from pkg_resources import iter_entry_points

    opts = copy.deepcopy(opts)
    cfg = read_setupcfg(opts['project']).to_dict()
    if 'pyscaffold' not in cfg:
        raise PyScaffoldTooOld
    pyscaffold = cfg['pyscaffold']
    metadata = cfg['metadata']
    # This would be needed in case of inplace updates, see issue #138, v4
    # if opts['project'] == '.':
    #   opts['project'] = metadata['name']
    # Overwrite only if user has not provided corresponding cli argument
    opts.setdefault('package', pyscaffold['package'])
    opts.setdefault('author', metadata['author'])
    opts.setdefault('email', metadata['author-email'])
    opts.setdefault('url', metadata['url'])
    opts.setdefault('description', metadata['description'])
    opts.setdefault('license', best_fit_license(metadata['license']))
    # Additional parameters compare with `get_default_options`
    opts['classifiers'] = metadata['classifiers'].strip().split('\n')
    # complement the cli extensions with the ones from configuration
    if 'extensions' in pyscaffold:
        cfg_extensions = pyscaffold['extensions'].strip().split('\n')
        opt_extensions = [ext.name for ext in opts['extensions']]
        add_extensions = set(cfg_extensions) - set(opt_extensions)
        for extension in iter_entry_points('pyscaffold.cli'):
            if extension.name in add_extensions:
                extension_obj = extension.load()(extension.name)
                if extension.name in pyscaffold:
                    ext_value = pyscaffold[extension.name]
                    extension_obj.args = ext_value
                    opts[extension.name] = ext_value
                opts['extensions'].append(extension_obj)
    return opts


def best_fit_license(txt):
    """Finds proper license name for the license defined in txt

    Args:
        txt (str): license name

    Returns:
        str: license name
    """
    ratings = {lic: levenshtein(txt, lic.lower()) for lic in licenses}
    return min(ratings.items(), key=itemgetter(1))[0]



# -*- coding: utf-8 -*-
"""
Extension that omits the creation of file `skeleton.py`
"""

from pathlib import PurePath as Path

from ..api import Extension, helpers


class NoSkeleton(Extension):
    """Omit creation of skeleton.py and test_skeleton.py"""
    def activate(self, actions):
        """Activate extension

        Args:
            actions (list): list of actions to perform

        Returns:
            list: updated list of actions
        """
        return self.register(
            actions,
            self.remove_files,
            after='define_structure')

    def remove_files(self, struct, opts):
        """Remove all skeleton files from structure

        Args:
            struct (dict): project representation as (possibly) nested
                :obj:`dict`.
            opts (dict): given options, see :obj:`create_project` for
                an extensive list.

        Returns:
            struct, opts: updated project representation and options
        """
        # Namespace is not yet applied so deleting from package is enough
        file = Path(opts['project'], 'src', opts['package'], 'skeleton.py')
        struct = helpers.reject(struct, file)
        file = Path(opts['project'], 'tests', 'test_skeleton.py')
        struct = helpers.reject(struct, file)
        return struct, opts



import numpy as np
import numbers
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import autodiffcst.AD as ad
import autodiffcst.admath as admath


class VAD():

    def __init__(self, val, der=None, der2=None, order=2, higher=None):
        """
        Overwrites the __init__ dunder method to create a new VAD object with initial value and derivatives.
    
                Parameters:
                        val (int or float or list or np.array): the initial value of the new VAD object.
                        der (int or float or list or np.array): first-order derivatives of the new AD object. 
                        der2 (int or float or list or np.array): second-order derivatives of the new AD object. 

                Returns:
                        None, but initializes AD object(s) when called
                
                Example:
                >>> x, y = VAD([1,2]) 
                >>> x
                AD(value: [1], derivatives: [1., 0.])

                >>> fs = VAD([1,2]) 
                >>> fs
                VAD(value: [1 2], derivatives: [[1. 0.]
                                                [0. 1.]])
        """
        self.val = np.array(val)
        if der is None:
            self.der = np.eye(len(self))
            self.der2 = np.zeros((len(self),len(self),len(self)))
        else:
            self.der = der
            self.der2 = der2
        self.tag = np.array([i for i in range(len(self))])
        self.size = len(self)

        if not isinstance(order, numbers.Integral):
            raise TypeError("Highest order of derivatives must be a positive integer.")
        elif order < 1:
            raise ValueError("Highest order of derivatives must be at least 1.")
        elif order > 2 and self.size > 1:
            raise Exception("We cannot handle derivatives of order > 2 for more than one scalar variables.")

        self.order = order
        self.higher = higher

        if self.size > 1:
            arr_ad = np.array([None]*len(self))
            for i in range(len(self)):
                arr_ad[i] = ad.AD(val=self.val[i], tag=self.tag[i], size = self.size,
                                der=self.der[i], der2=self.der2[i])
            self.variables = arr_ad
        else:
            arr_ad = np.array([None]*len(self))
            arr_ad[0] = ad.AD(val=self.val[0], tag=self.tag[0], size = 1,
                                der=self.der[0], der2=self.der2[0], 
                                order=self.order, higher = self.higher)
            self.variables = arr_ad

    def __str__(self):
        """
        Overwrites the __str__ dunder method to nicely turn an AD object into a string.
    
                Parameters:
                        self (AD): the AD object that __str__ is called upon.
    
                Returns:
                        A string containing the current value and derivatives of the AD object.
                
                Example:
                >>> fs = VAD([1,2]) 
                >>> print(fs)
                VAD(value: [1 2], derivatives: [[1. 0.]
                                                [0. 1.]])
        """
        return "VAD(value: {0}, derivatives: {1})".format(self.val, self.der)

    def __repr__(self):
        """
        Overwrites the __repr__ dunder method to nicely print an AD object.
    
                Parameters:
                        self (AD): the AD object that __repr__ is called upon.
    
                Returns:
                        A string containing the current value and derivatives of the AD object.

                Example:
                >>> fs = VAD([1,2]) 
                >>> fs
                VAD(value: [1 2], derivatives: [[1. 0.]
                                                [0. 1.]])
        
        """
        return "VAD(value: {0}, derivatives: {1})".format(self.val, self.der)

    def __len__(self):
        """
        Overwrites the __len__ dunder method to get the length of the VAD object's values, namely the number of variables.
    
                Parameters:
                        self (VAD): the VAD object that __len__ is called upon.
    
                Returns:
                        An integer representing the length of the vector dimension that the VAD object resides in.

                Example:
                >>> len(VAD([1,2]))
                2
        """    
        try:
            return len(self.val)
        except TypeError:
            return 1

    ## getter
    def __getitem__(self, pos):
        """
        Overwrites the __getitem__ dunder method to get variable at certain position of the VAD object.
    
                Parameters:
                        self (VAD): the VAD object that __getitem__ is called upon.
                        pos (int): a valid position of a variable.
    
                Returns:
                        AD object at the position

                Example:
                >>> VAD([1,2])[0]
                AD(value: [1], derivatives: [1. 0.])
        """  
        return self.variables[pos]


    ## setter, Do we want this?
    def __setitem__(self, pos, newAD):
        """
        Overwrites the __eq__ dunder method to check if two VAD objects are equal in value.
    
                Parameters:
                        self (VAD): the VAD object that __eq__ is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        True if all the values of two VAD objects are equal.
                        False if otherwise.

                Example:
                >>> VAD([1,2]) == VAD([1,3])
                False
        """
        self.variables[pos] = newAD
        
    # Comparison Equal
    def __eq__(self, other):
        """
        Overwrites the __eq__ dunder method to check if two VAD objects are equal in value.
    
                Parameters:
                        self (VAD): the VAD object that __ne__ is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        False if all the values of two VAD objects are equal.
                        True if otherwise.

                Example:
                >>> VAD([1,2]) != VAD([1,3])
                True
        """
        if isinstance(other, VAD):
            if np.sum(self.val == other.val) == len(self): 
                return True
                
            else:
                return False
        else:
            raise TypeError("Invalid Comparison. VAD object can only be compared with VAD.")

    def __ne__(self, other):
        """
        Overwrites the __ne__ dunder method to check if two VAD objects are equal in value.
    
                Parameters:
                        self (VAD): the VAD object that __ne__ is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        False if all the values of two VAD objects are equal.
                        True if otherwise.

                Example:
                >>> VAD([1,2]) != VAD([1,3])
                True
        """
        return not self == other

    def isequal(self, other):
        """
        Compare value of two VAD objects' values element wise.

                Parameters:
                        self (VAD): the VAD object that isequal is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        an array of True or False depends on the comparison result.
        
                Example:
                >>> a = VAD([1,2,3])
                >>> b = VAD([2,2,3])
                >>> a.isequal(b)
                array([False,True,True]) 
        """
        if isinstance(other, VAD):
            return self.val == other.val      
        else:
            raise TypeError("The input must also be a VAD object.")
    
    def fullequal(self, other):
        """
         Compare value, first derivatives and second derivatives of two VAD objects.

                Parameters:
                        self (VAD): the VAD object that fullequal is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        True if everything is equal.
                        False otherwise.
     
                Example:
                >>> a = VAD([1,2,3])
                >>> b = VAD([1,2,3]) * 1
                >>> a.fullequal(b)
                True
        """
        if isinstance(other, VAD):
            return np.allclose(self.val, other.val) and np.allclose(self.der,other.der) and np.allclose(self.der2, other.der2)
        else:
            raise TypeError("Invalid Comparison. VAD object can only be compared with VAD.")


    def __gt__(self, other):
        """
        Overwrites the __gt__ dunder method to compare value of two VAD objects.

                Parameters:
                        self (VAD): the VAD object that __gt__ is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        True if self.val is larger than other.val elementwise.
                        False otherwise.
                Example:
                >>> a = VAD([2,2,3])
                >>> b = VAD([1,2,2])
                >>> a > b
                False
        """
        if isinstance(other, VAD):
            return np.sum(self.val > other.val) == len(self)
        else:
            raise TypeError("Invalid Comparison. VAD object can only be compared with VAD.")



    def __ge__(self, other):
        """
        Overwrites the __ge__ dunder method to compare value of two VAD objects.

                Parameters:
                        self (VAD): the VAD object that __ge__ is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        True if self.val is greater than or equal to other.val elementwise.
                        False otherwise.
                Example:
                >>> a = VAD([2,2,3])
                >>> b = VAD([1,2,2])
                >>> a >= b
                True
        """
        if isinstance(other, VAD):
            return np.sum(self.val >= other.val) == len(self)
        else:
            raise TypeError("Invalid Comparison. VAD object can only be compared with VAD.")
            
    def isgreater(self, other):
        """
        Compare value of two VAD objects element wise

                Parameters:
                        self (VAD): the VAD object that isgreater is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        an array of boolean values.

                Example:
                >>> a = VAD([2,2,3])
                >>> b = VAD([1,2,2])
                >>> a.isgreater(b)
                array([ True, False,  True])
        
        """
        if isinstance(other, VAD):
            return self.val > other.val      
        else:
            raise TypeError("The input must also be a VAD object.")


    def __lt__(self, other):
        """
        Overwrites the __lt__ dunder method to compare value of two VAD objects.


                Parameters:
                        self (VAD): the VAD object that __lt__ is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        True if all values in self is less than values in other.
                        False otherwise.
                        
                Example:
                >>> a = VAD([2,2,3])
                >>> b = VAD([1,2,2])
                >>> a < b
                False
        """
        if isinstance(other, VAD):
            return np.sum(self.val < other.val) == len(self)
        else:
            raise TypeError("Invalid Comparison. VAD object can only be compared with VAD.")


    def __le__(self, other):
        """
        Overwrites the __le__ dunder method to compare value of two VAD objects.


                Parameters:
                        self (VAD): the VAD object that __le__ is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        True if all values in self is less than or equal to values in other.
                        False otherwise.
                        
                Example:
                >>> a = VAD([2,2,3])
                >>> b = VAD([1,2,2])
                >>> a <= b
                False
        """
        if isinstance(other, VAD):
            return np.sum(self.val <= other.val) == len(self)
        else:
            raise TypeError("Invalid Comparison. VAD object can only be compared with VAD.")
            
    def isless(self, other):
        """
        Compare value of two VAD objects element wise

                Parameters:
                        self (VAD): the VAD object that isgreater is called upon.
                        other (VAD): the VAD object to be compared with.
    
                Returns:
                        an array of boolean values.

                Example:
                >>> a = VAD([2,2,3])
                >>> b = VAD([1,2,2])
                >>> a.isless(b)
                array([False, False, False])
        
        """
        if isinstance(other, VAD):
            return self.val < other.val      
        else:
            raise TypeError("The input must also be a VAD object.")
                
    ## Unary 
    def __neg__(self):
        """
        Overwrites the __neg__ dunder method to get the negation of the VAD object.
    
                Parameters:
                        self (VAD): the VAD object that __neg__ is called upon.
    
                Returns:
                        A new VAD object which has the negated value and derivative of the current VAD.

                Example:
                >>> -VAD([1,2])
                VAD(value: [-1,-2], derivatives: [[-1. -0.]
                                                  [-0. -1.]])
        """
        AD_result = self.variables * -1
        return set_VAD(AD_result) 

    ## Addition
    def __add__(self, other):
        """
        Overwrites the __add__ dunder method to apply addition to an VAD object.
    
                Parameters:
                        self (VAD): An VAD object to be applied addition to
                        other (VAD or valid input for the numpy operation): the object to be added to self
    
                Returns:
                        new_self (VAD): the new VAD object after applying addition

                Example:
                >>> a = VAD([1,2])
                >>> a + 3
                VAD(value: [4., 5.], derivatives: [[1., 0.],
                                                   [0., 1.]])
        """
        AD_result = self.variables + other
        return set_VAD(AD_result)  
        
    def __radd__(self, other):
        """
        Overwrites the __radd__ dunder method to apply addition to an VAD object 
        when the VAD object is on the right side of the addition sign.
    
                Parameters:
                        self (VAD): An AD object to be applied addition to
                        other (VAD or valid input for the numpy operation): the object to be added to self
    
                Returns:
                        new_self (VAD): the new VAD object after applying addition

                Example:
                >>> a = VAD([1,2])
                >>> 3 + a
                VAD(value: [4., 5.], derivatives: [[1., 0.],
                                                   [0., 1.]])
        """          
        return self + other
    
    def __iadd__(self, other):
        """
        Overwrites the __iadd__ dunder method to apply addition to an VAD object when the operation "+=" is used.
    
                Parameters:
                        self (VAD): An VAD object to be applied addition to
                        other (VAD or valid input for the numpy operation): the object to be added to self
    
                Returns:
                        None, but a new VAD will be assigned to the original variable.
                
                Example:
                >>> a = VAD([1,2])
                >>> a += 1
                >>> a
                VAD(value: [2., 3.], derivatives: [[1., 0.],
                                                   [0., 1.]])
        """         
        return self + other

    # Division
    def __sub__(self, other):
        """
        Overwrites the __sub__ dunder method to apply substraction to an VAD object.
    
                Parameters:
                        self (VAD): An VAD object to be applied substraction to
                        other (VAD or valid input for the numpy operation): the object to be substracted to self
    
                Returns:
                        new_self (VAD): the new VAD object after applying substraction

                Example:
                >>> a = VAD([1,2])
                >>> a - 3
                VAD(value: [-2., -1.], derivatives: [[1., 0.],
                                                     [0., 1.]])
        """       
        return self + (-1)*other
    
    def __rsub__(self, other):
        """
        Overwrites the __rsub__ dunder method to apply substraction to an VAD object.
    
                Parameters:
                        self (VAD): An VAD object to be applied substraction to
                        other (VAD or valid input for the numpy operation): the object to be substracted to self
    
                Returns:
                        new_self (VAD): the new VAD object after applying substraction

                Example:
                >>> a = VAD([1,2])
                >>> 3 - a
                VAD(value: [2., 1.], derivatives: [[-1., -0.],
                                                   [-0., -1.]])
        """  
        return (-1)*self + other
    
    def __isub__(self, other):
        """
        Overwrites the __isub__ dunder method to apply substraction to an VAD object.
    
                Parameters:
                        self (VAD): An VAD object to be applied substraction to
                        other (VAD or valid input for the numpy operation): the object to be substracted to self
    
                Returns:
                        None, but a new VAD will be assigned to the original variable.
                
                Example:
                >>> a = VAD([1,2])
                >>> a -= 3
                >>> a
                VAD(value: [-2., -1.], derivatives: [[1., 0.],
                                                     [0., 1.]])
        """            
        return self - other


    ## Mod
    def __mod__(self, other):
        """
        Overwrites the __mod__ dunder method to apply mod to an VAD object.
    
                Parameters:
                        self (VAD): An VAD object to be applied mod to
                        other (int or float): the number that self is modded by
    
                Returns:
                        new_val: the value of the VAD object after applying mod
                
                Example:
                >>> a = VAD([1,2])
                >>> a % 2
                array([1, 0])

        """          
        if isinstance(other, int) or isinstance(other, float):
            new_val = self.val % other
            print("Warning: the mod function does not yield any derivatives. Instead, the function applies mod to the current value of the AD object and returns the result.")
            return new_val
        else:
            raise TypeError("You can only mode by an integer or a float.")
    
    def __imod__(self, other):
        """
        Overwrites the __imod__ dunder method to apply mod to an VAD object when the operation "%=" is used.
    
                Parameters:
                        self (VAD): An VAD object to be applied mod to
                        other (int or float): the number that self is modded by
    
                Returns:
                        None, but an array of mod values will be assigned to the original variable.
                
                Example:
                >>> a = VAD([1,2])
                >>> a %= 2
                >>> a
                array([1, 0])

        """           
        return self % other    

    ## Multiplication
    def __mul__(self, other):
        """
        Overwrites the __mul__ dunder method to apply multiplication to an VAD object.
    
                Parameters:
                        self (VAD): An AD object to be applied multiplication to
                        other (VAD or valid input for the numpy operation): the object to be multiplied to self
    
                Returns:
                        new_self (VAD): the new VAD object after applying multiplication
                
                Example:
                >>> a = VAD([1,2])
                >>> a*2
                VAD(value: [2., 4.], derivatives: [[2., 0.],
                                                   [0., 2.]])
        """     
        AD_result = self.variables * other
        return set_VAD(AD_result)    

    def __rmul__(self, other):
        """
        Overwrites the __rmul__ dunder method to apply multiplication to an VAD object.
    
                Parameters:
                        self (VAD): An AD object to be applied multiplication to
                        other (VAD or valid input for the numpy operation): the object to be multiplied to self
    
                Returns:
                        new_self (VAD): the new VAD object after applying multiplication
                
                Example:
                >>> a = VAD([1,2])
                >>> 2*a
                VAD(value: [2., 4.], derivatives: [[2., 0.],
                                                   [0., 2.]])
        """            
        return self * other

    def __imul__(self, other):
        """
        Overwrites the __rmul__ dunder method to apply multiplication to an VAD object.
    
                Parameters:
                        self (VAD): An AD object to be applied multiplication to
                        other (VAD or valid input for the numpy operation): the object to be multiplied to self
    
                Returns:
                        None, but a new VAD will be assigned to the original variable.
                
                Example:
                >>> a = VAD([1,2])
                >>> a *= 2
                >>> a
                VAD(value: [2., 4.], derivatives: [[2., 0.],
                                                   [0., 2.]])
        """          
        return self * other
    
    
    ## Division
    def __truediv__(self, other):
        """
        Overwrites the __truediv__ dunder method to apply division to an VAD object.
    
                Parameters:
                        self (VAD): An VAD object to be applied division to
                        other (VAD or valid input for the numpy operation): the object that self is divided by
    
                Returns:
                        new_self (VAD): the new VAD object after applying division
                        
                Example:
                >>> a = VAD([2,4])
                >>> a / 2
                VAD(value: [1., 2.], derivatives: [[0.5, 0.],
                                                   [0., 0.5]])
        """
        AD_result = self.variables / other
        return set_VAD(AD_result)   
        
    
    def __rtruediv__(self, other):
        """
        Overwrites the __rtruediv__ dunder method to apply division to an AD object 
        when the AD object is on the right side of the division sign.
    
                Parameters:
                        self (AD): An AD object to be applied division to
                        other (VAD or valid input for the numpy operation): the object that self is divided by
    
                Returns:
                        new_self (AD): the new AD object after applying division
                
                Example:
                >>> a = VAD([2,4])
                >>> 2 / a
                VAD(value: [1., 0.5], derivatives: [[-0.5, 0.],
                                                   [0., -0.125]])
        """
        AD_result = other/self.variables
        return set_VAD(AD_result)
        
    
    def __itruediv__(self, other):
        """
        Overwrites the __itruediv__ dunder method to apply division to an VAD object when the operation "/=" is used.
    
                Parameters:
                        self (VAD): An VAD object to be applied division to
                        other (VAD or valid input for the numpy operation): the object that self is divided by
    
                Returns:
                        None, but a new VAD will be assigned to the original variable.
                
                Example:
                >>> a = VAD([2,4])
                >>> a /= 2
                >>> a
                VAD(value: [1., 2.], derivatives: [[0.5, 0.],
                                                   [0., 0.5]])
                    
        """          
        return self / other
    
    
  ## power
    
    def __pow__(self, other):
        """
        Overwrites the __pow__ dunder method to apply power function to an VAD object.
    
                Parameters:
                        self (VAD): An VAD object to be applied power function to
                        other (VAD or valid input for the numpy operation): the object that self's power will be raised to
    
                Returns:
                        new_self (VAD): the new VAD object after applying power function
                
                Example:
                >>> a = VAD([1,2])
                >>> a ** 2
                VAD(value: [1., 4.], derivatives: [[2., 0.],
                                                   [0., 4.]])
        """           
        AD_result = self.variables ** other
        return set_VAD(AD_result)  

    
    def __ipow__(self, other):
        """
        Overwrites the __ipow__ dunder method to apply power function to an VAD object when the operation "**=" is used.
    
                Parameters:
                        self (VAD): An VAD object to be applied power function to
                        other (VAD or valid input for the numpy operation): the object that self's power will be raised to
    
                Returns:
                        None, but a new VAD will be assigned to the original variable.

                Example:
                >>> a = VAD([1,2])
                >>> a **= 2
                >>> a
                VAD(value: [1., 4.], derivatives: [[2., 0.],
                                                   [0., 4.]])
        """           
        return self ** other

    
    def __rpow__(self, other):
        """
        Overwrites the __rpow__ dunder method to apply power function to an VAD object 
        when the VAD object is on the right side of the power sign.
    
                Parameters:
                        self (VAD): An VAD object to be applied power function to
                        other (VAD or valid input for the numpy operation): the object that self's power will be raised to
    
                Returns:
                        new_self (VAD): the new VAD object after applying power function
        
                Example:
                >>> a = VAD([1,3])
                >>> 2 ** a
                VAD(value: [2., 8.], derivatives: [[2.*np.log(2.), 0.],
                                                   [0., 8.*np.log(2.)]])
        """
        AD_result = other ** self.variables
        return set_VAD(AD_result)


    def diff(self, direction, order = 1):
        """
        Calculate and return the derivatives of the function represented by an VAD object.
    
                Parameters:
                        self (VAD): the VAD object whose derivatives will be calculated.
                        direction (list): the seed indicating which variable's derivative should be returned
                        order (1 or 2): the order of the derivatives
                Returns:
                        An array representing the derivatives (or derivative) of the VAD object, 
                        with directions indicted by the input direction
                
                Example:
                >>> a = VAD([3,1])
                >>> f = 2 * a
                >>> f.diff(1,1)
                np.array([0.,2.])             
        """    
        if order == 1 and isinstance(direction,int):
            return self.der[:,direction]
                
        elif order == 2 and isinstance(direction, list) and len(direction) ==2:
            return self.der2[:,direction[0],direction[1]]
        else:
            raise Exception("Order exceeds 2 or length of direction and order don't match.")

# helper function
def set_VAD(ADs):
    """
        Create a VAD object with a list of AD objects.
    
                Parameters:
                        ADs (array or np.array): a list or np.array of AD objects
                Returns:
                        A VAD object that has the values and derivatives of the list of AD objects
                
                Example:
                >>> x = ad.AD(1, tag=0, size=2)
                >>> y = ad.AD(2, tag=1, size=2)
                >>> ADs = np.array([x, y])
                >>> set_VAD(ADs) 
                VAD([1,2])       
    """ 
    new_val = np.concatenate([ADs[i].val for i in range(len(ADs))])
    new_der = np.array([ADs[i].der for i in range(len(ADs))])
    new_der2 = np.array([ADs[i].der2 for i in range(len(ADs))])
    return VAD(new_val, new_der, new_der2)

def my_decorator(func):
    """
        Helper function that enables directly using functions from admath.
    
                Parameters:
                        func: the function to be applied
                Returns:
                        A VAD object after the function is applied to it
                
                Example:
                >>> func = my_decorator(admath.exp)
                >>> y = VAD([1, 2, 3])
                >>> func(y)
                exp(y)
    """ 
    def wrapper(vad):
        try:
            AD_result = np.array([func(ad) for ad in vad.variables])
            return set_VAD(AD_result)
        except:
            return func(vad)
    return wrapper


exp = my_decorator(admath.exp)
abs = my_decorator(admath.abs)
log = my_decorator(admath.log)
sqrt = my_decorator(admath.sqrt)
sin = my_decorator(admath.sin)
cos = my_decorator(admath.cos)
tan = my_decorator(admath.tan)
sinh = my_decorator(admath.sinh)
cosh = my_decorator(admath.cosh)
tanh = my_decorator(admath.tanh)


def pow(vad,y):
    """
        Do the power operation on VAD objects, functions similar to **.
    
                Parameters:
                        vad (VAD): the VAD object to apply power on
                        y (VAD or valid input for the numpy operation): the object that vad's power will be raised to
    
                Returns:
                        The new VAD object after applying power function

                Example:
                >>> a = VAD([1,2])
                >>> pow(a, 2)
                VAD(value: [1., 4.], derivatives: [[2., 0.],
                                                   [0., 4.]])
    """ 
    try:
        AD_result = np.array([ad**y for ad in vad.variables])
        return set_VAD(AD_result)
    except:
        return np.power(vad,y)



# jacobian
def jacobian(funcs):
    """
        Return the Jacobian matrix of the input function(s).
    
                Parameters:
                        funcs (VAD or list of VAD): the VAD object(s) for which we want to calculate the Jacobian of
    
                Returns:
                        The Jacobian matrix of the input function(s)

                Example:
                >>> x = VAD([3, 1])
                >>> f = 2 * x
                >>> jacobian(f),np.array([[2., 0.],[0.,2.]]))
    """
    diffs = []
    if isinstance(funcs,VAD) or isinstance(funcs,ad.AD):
        return funcs.der
    elif isinstance(funcs,list) or isinstance(funcs,np.ndarray):
        for func in funcs:
            if not isinstance(func, ad.AD):
                raise TypeError("Invalid Type. All functions should be AD object.")
            diffs.append(func.der)
        return np.vstack(diffs)

# hessian
def hessian(func):
    """
    Return the Hessian matrix of the input function.
    
            Parameters:
                    func (AD): the AD object for which we want to calculate the Hessian of
    
            Returns:
                    The Hessian matrix of the input function

            Raise:
                    TypeError if the input is VAD and has multiple functions

            Example:
            >>> x = VAD([3, 1])
            >>> g = 2 * x[0]
            >>> hessian(g)
            np.array([[0., 0.], [0., 0.]]))
    """ 
    if isinstance(func, VAD):
        raise TypeError("Invalid Type. Sorry, we cannot handle multiple functions for Hessian.")
    elif isinstance(func, ad.AD):
        hessian = func.der2
        return hessian
    else:
        raise TypeError("Invalid Type. Function should be an AD object.")






# -*- coding: utf-8 -*-
"""
Functionality to generate and work with the directory structure of a project
"""

import os
from os.path import join as join_path

from . import templates, utils


class FileOp(object):
    """Namespace for file operations during an update"""

    NO_OVERWRITE = 0
    """Do not overwrite an existing file during update
    (still created if not exists)
    """

    NO_CREATE = 1
    """Do not create the file during an update"""


def define_structure(_, opts):
    """Creates the project structure as dictionary of dictionaries

    Args:
        struct (dict): previous directory structure (ignored)
        opts (dict): options of the project

    Returns:
        tuple(dict, dict):
            structure as dictionary of dictionaries and input options
    """
    struct = {opts['project']: {
        '.gitignore': (templates.gitignore(opts), FileOp.NO_OVERWRITE),
        'src': {
            opts['package']: {'__init__.py': templates.init(opts),
                              'skeleton.py': (templates.skeleton(opts),
                                              FileOp.NO_CREATE)},
        },
        'tests': {'conftest.py': (templates.conftest_py(opts),
                                  FileOp.NO_OVERWRITE),
                  'test_skeleton.py': (templates.test_skeleton(opts),
                                       FileOp.NO_CREATE)},
        'docs': {'conf.py': templates.sphinx_conf(opts),
                 'authors.rst': templates.sphinx_authors(opts),
                 'index.rst': (templates.sphinx_index(opts),
                               FileOp.NO_OVERWRITE),
                 'license.rst': templates.sphinx_license(opts),
                 'changelog.rst': templates.sphinx_changelog(opts),
                 'Makefile': templates.sphinx_makefile(opts),
                 '_static': {
                     '.gitignore': templates.gitignore_empty(opts)}},
        'README.rst': (templates.readme(opts), FileOp.NO_OVERWRITE),
        'AUTHORS.rst': (templates.authors(opts), FileOp.NO_OVERWRITE),
        'LICENSE.txt': (templates.license(opts), FileOp.NO_OVERWRITE),
        'CHANGELOG.rst': (templates.changelog(opts), FileOp.NO_OVERWRITE),
        'setup.py': templates.setup_py(opts),
        'setup.cfg': (templates.setup_cfg(opts), FileOp.NO_OVERWRITE),
        'requirements.txt': (templates.requirements(opts),
                             FileOp.NO_OVERWRITE),
        '.coveragerc': (templates.coveragerc(opts), FileOp.NO_OVERWRITE)}}

    return struct, opts


def create_structure(struct, opts, prefix=None):
    """Manifests a directory structure in the filesystem

    Args:
        struct (dict): directory structure as dictionary of dictionaries
        opts (dict): options of the project
        prefix (str): prefix path for the structure

    Returns:
        tuple(dict, dict):
            directory structure as dictionary of dictionaries (similar to
            input, but only containing the files that actually changed) and
            input options

    Raises:
        :obj:`RuntimeError`: raised if content type in struct is unknown
    """
    update = opts.get('update') or opts.get('force')
    pretend = opts.get('pretend')

    if prefix is None:
        prefix = os.getcwd()

    changed = {}

    for name, content in struct.items():
        if isinstance(content, str):
            utils.create_file(join_path(prefix, name), content, pretend)
            changed[name] = content
        elif isinstance(content, dict):
            utils.create_directory(join_path(prefix, name), update, pretend)
            changed[name], _ = create_structure(
                    struct[name], opts, prefix=join_path(prefix, name))
        elif content is None:
            pass
        else:
            raise RuntimeError("Don't know what to do with content type "
                               "{type}.".format(type=type(content)))

    return changed, opts
